""" The central part of the tool: Core """
import datetime

import fnmatch
import glob
import importlib as il
import json
import logging
import os
import shutil
import socket
import tempfile
import time
import copy
import sys
import platform

import yaml
from builtins import str

from yandextank.common.const import RetCode
from yandextank.common.exceptions import GeneratorNotFound, PluginNotPrepared
from yandextank.common.interfaces import GeneratorPlugin, MonitoringPlugin, MonitoringDataListener
from yandextank.plugins.DataUploader.client import LPRequisites
from yandextank.validator.validator import TankConfig, ValidationError
from yandextank.aggregator import TankAggregator
from yandextank.aggregator.aggregator import DataPoller
from yandextank.common.util import pid_exists

from yandextank.contrib.netort.netort.resource import manager as default_resource_manager
from yandextank.contrib.netort.netort.process import execute
from yandextank.version import VERSION

logger = logging.getLogger(__name__)

CONFIGINITIAL = 'configinitial.yaml'
VALIDATED_CONF = 'validated_conf.yaml'

CLOUD_KEY = 'cloud_job_id'
TANK_KEY = 'tank_job_id'
JOBS_STORAGE_FILE_ENV = 'JOBS_STORAGE_FILE'
DEFAULT_JOBS_STORAGE_FILE = '/tmp/yandex-tank/storage.data'


class Job(object):
    def __init__(
            self,
            monitoring_plugins,
            aggregator,
            tank,
            generator_plugin=None):
        """

        :type aggregator: TankAggregator
        :type monitoring_plugins: list of
        """
        self.monitoring_plugins = monitoring_plugins
        self.aggregator = aggregator
        self.tank = tank
        self._phantom_info = None
        self.generator_plugin = generator_plugin

    def subscribe_plugin(self, plugin):
        self.aggregator.add_result_listener(plugin)
        for monitoring_plugin in self.monitoring_plugins:
            monitoring_plugin.add_listener(plugin)

    @property
    def phantom_info(self):
        if self._phantom_info is None:
            raise PluginNotPrepared
        return self._phantom_info

    @phantom_info.setter
    def phantom_info(self, info):
        self._phantom_info = info


def parse_plugin(s):
    try:
        plugin, config_section = s.split()
    except ValueError:
        plugin, config_section = s, None
    return plugin, config_section


class LockError(Exception):
    pass


class TankCore(object):
    SECTION = 'core'
    SECTION_META = 'meta'
    PLUGIN_PREFIX = 'plugin_'
    PID_OPTION = 'pid'
    UUID_OPTION = 'uuid'
    API_JOBNO = 'api_jobno'

    def __init__(
        self,
        configs,
        interrupted_event,
        info,
        storage=None,
        skip_base_cfgs=True,
        resource_manager=None,
        plugins_implicit_enabling=False,
    ):
        """
        :param configs: list of dict
        :param interrupted_event: multiprocessing.Event
        :type info: yandextank.common.interfaces.TankInfo
        """
        self.output = {}
        self.raw_configs = configs
        self.info = info
        self._plugins = None
        self._artifacts_dir = None
        self.artifact_files = {}
        self.artifacts_to_send = []
        self._artifacts_base_dir = None
        self.manual_start = False
        self.scheduled_start = None
        self.taskset_path = None
        self.taskset_affinity = None
        self._job = None
        self._cfg_snapshot = None
        self._data_poller = None
        self._extra_plugins = []

        self.interrupted = interrupted_event
        self.resource_manager = resource_manager or default_resource_manager

        self.error_log = None
        self.monitoring_data_listeners = []

        error_output = 'validation_error.yaml'
        self.config, self.configinitial = TankConfig(self.raw_configs,
                                                     with_dynamic_options=True,
                                                     core_section=self.SECTION,
                                                     error_output=error_output,
                                                     skip_base_cfgs=skip_base_cfgs,
                                                     plugins_implicit_enabling=plugins_implicit_enabling).validate()

        self.test_id = self.get_option(self.SECTION, 'artifacts_dir',
                                       datetime.datetime.now().strftime("%Y-%m-%d_%H-%M-%S.%f"))
        self.lock_dir = self.get_option(self.SECTION, 'lock_dir')
        self.aggregator_max_termination_timeout = self.get_option(self.SECTION, 'aggregator_max_termination_timeout', 60)
        self.skip_generator_check = self.get_option(self.SECTION, 'skip_generator_check', False)
        with open(os.path.join(self.artifacts_dir, CONFIGINITIAL), 'w') as f:
            yaml.dump(self.configinitial, f)
        self.add_artifact_file(error_output)
        self.add_artifact_to_send(LPRequisites.CONFIGINITIAL, yaml.dump(self.configinitial))
        configinfo = self.config.validated.copy()
        configinfo.setdefault(self.SECTION, {})
        configinfo[self.SECTION][self.API_JOBNO] = self.test_id
        self.add_artifact_to_send(LPRequisites.CONFIGINFO, yaml.dump(configinfo))
        with open(os.path.join(self.artifacts_dir, VALIDATED_CONF), 'w') as f:
            yaml.dump(configinfo, f)
        logger.info('New test id %s' % self.test_id)
        self.storage = storage or JobsStorage()
        self.errors = []

    @property
    def cfg_snapshot(self):
        if not self._cfg_snapshot:
            self._cfg_snapshot = str(self.config)
        return self._cfg_snapshot

    @staticmethod
    def get_available_options():
        # todo: should take this from schema
        return [
            "artifacts_base_dir", "artifacts_dir",
            "taskset_path", "affinity"
        ]

    @property
    def plugins(self):
        """
        :returns: {plugin_name: plugin_class, ...}
        :rtype: dict
        """
        if self._plugins is None:
            self.load_plugins()
            if self._plugins is None:
                self._plugins = {}
        return self._plugins

    @property
    def artifacts_base_dir(self):
        if not self._artifacts_base_dir:
            try:
                artifacts_base_dir = os.path.abspath(self.get_option(self.SECTION, "artifacts_base_dir"))
            except ValidationError:
                artifacts_base_dir = os.path.abspath('logs')
            if not os.path.exists(artifacts_base_dir):
                os.makedirs(artifacts_base_dir)
                os.chmod(self.artifacts_base_dir, 0o755)
            self._artifacts_base_dir = artifacts_base_dir
        return self._artifacts_base_dir

    @property
    def data_poller(self):
        if self._data_poller is None:
            self._data_poller = DataPoller(
                poll_period=0.5,
                max_wait=self.get_option(self.SECTION, "aggregator_max_wait"),
            )
        return self._data_poller

    def load_plugins(self):
        """
        Tells core to take plugin options and instantiate plugin classes
        """
        logger.info("Loading plugins...")
        generators_counter = 0
        for (plugin_name, plugin_path, plugin_cfg) in self.config.plugins:
            logger.debug("Loading plugin %s from %s", plugin_name, plugin_path)
            if plugin_path == "yandextank.plugins.Overload":
                logger.warning(
                    "Deprecated plugin name: 'yandextank.plugins.Overload'\n"
                    "There is a new generic plugin now.\n"
                    "Correcting to 'yandextank.plugins.DataUploader overload'")
                plugin_path = "yandextank.plugins.DataUploader overload"
            try:
                logger.info("Trying to import plugin %s from path %s", plugin_name, plugin_path)
                plugin = il.import_module(plugin_path)
            except ImportError:
                logger.exception('Plugin name %s path %s import error', plugin_name, plugin_path)
                raise
            try:
                instance = getattr(plugin, 'Plugin')(self, cfg=plugin_cfg, name=plugin_name)
                if isinstance(instance, GeneratorPlugin):
                    generators_counter += 1
            except AttributeError:
                logger.warning('Plugin %s classname should be `Plugin`', plugin_name)
                raise
            else:
                self.register_plugin(self.PLUGIN_PREFIX + plugin_name, instance)
        for plugin_name, factory in self._extra_plugins:
            self.register_plugin(self.PLUGIN_PREFIX + plugin_name, factory(self))
        if not generators_counter and not self.skip_generator_check:
            raise GeneratorNotFound('Generator plugin is missing in test config. '
                                    'Enable load generator plugin or use "skip_generator_check" '
                                    'core option to run without load generator')
        logger.debug("Plugin instances: %s", self._plugins)

    @property
    def job(self):
        if not self._job:
            # monitoring plugin
            monitorings = [plugin for plugin in self.plugins.values() if isinstance(plugin, MonitoringPlugin)]
            # generator plugin
            try:
                gen = self.get_plugin_of_type(GeneratorPlugin)
            except KeyError:
                logger.warning("Load generator not found")
                gen = GeneratorPlugin(self, {}, 'generator dummy')
            # aggregator
            aggregator = TankAggregator(gen, self.data_poller, termination_timeout=self.aggregator_max_termination_timeout)
            self._job = Job(monitoring_plugins=monitorings,
                            generator_plugin=gen,
                            aggregator=aggregator,
                            tank=socket.getfqdn())
        return self._job

    def plugins_configure(self):
        """        Call configure() on all plugins        """
        self.publish("core", "stage", "configure")

        logger.info("Configuring plugins...")
        self.taskset_affinity = self.get_option(self.SECTION, 'affinity')
        if self.taskset_affinity:
            self.__setup_taskset(self.taskset_affinity, pid=os.getpid())

        for plugin in self.plugins.values():
            if not self.interrupted.is_set():
                logger.debug("Configuring %s", plugin)
                plugin.configure()
                if isinstance(plugin, MonitoringDataListener):
                    self.monitoring_data_listeners.append(plugin)

    def register_external_plugin(self, name: str, factory):
        self._extra_plugins.append((name, factory))

    def plugins_prepare_test(self):
        """ Call prepare_test() on all plugins        """
        logger.info("Preparing test...")
        self.publish("core", "stage", "prepare")
        for plugin in self.plugins.values():
            if not self.interrupted.is_set():
                logger.debug("Preparing %s", plugin)
                plugin.prepare_test()

    def plugins_start_test(self):
        """        Call start_test() on all plugins        """
        if not self.interrupted.is_set():
            logger.info("Starting test...")
            self.publish("core", "stage", "start")
            self.job.aggregator.start_test()
            for plugin_name, plugin in self.plugins.items():
                logger.debug("Starting %s", plugin)
                start_time = time.time()
                plugin.start_test()
                logger.info("Plugin {0:s} required {1:f} seconds to start".format(plugin_name,
                                                                                  time.time() - start_time))
            self.publish('generator', 'test_start', self.job.generator_plugin.start_time)

    def wait_for_finish(self):
        """
        Call is_test_finished() on all plugins 'till one of them initiates exit
        """
        if not self.interrupted.is_set():
            logger.info("Waiting for test to finish...")
            logger.info('Artifacts dir: {dir}'.format(dir=self.artifacts_dir))
            self.publish("core", "stage", "shoot")
            if not self.plugins:
                raise RuntimeError("It's strange: we have no plugins loaded...")

        while not self.interrupted.is_set():
            begin_time = time.time()
            aggr_retcode = self.job.aggregator.is_test_finished()
            if aggr_retcode >= 0:
                return aggr_retcode
            for plugin_name, plugin in self.plugins.items():
                logger.debug("Polling %s", plugin)
                try:
                    retcode = plugin.is_test_finished()
                    if retcode >= 0:
                        for e in plugin.errors:
                            self.errors.append(f'{plugin_name}: {e}')
                        return retcode
                except Exception:
                    logger.warning('Plugin {} failed:'.format(plugin_name), exc_info=True)
                    if isinstance(plugin, GeneratorPlugin):
                        return RetCode.ERROR
                    else:
                        logger.warning('Disabling plugin {}'.format(plugin_name))
                        plugin.is_test_finished = lambda: RetCode.CONTINUE
            end_time = time.time()
            diff = end_time - begin_time
            logger.debug("Polling took %s", diff)
            logger.debug("Tank status: %s", json.dumps(self.info.get_info_dict()))
            # screen refresh every 0.5 s
            if diff < 0.5:
                time.sleep(0.5 - diff)
        return 1

    def plugins_end_test(self, retcode):
        """        Call end_test() on all plugins        """
        logger.info("Finishing test with received RC: %s", retcode)
        self.publish("core", "stage", "end")
        self.publish('generator', 'test_end', time.time())

        logger.info("Stopping load generator and aggregator")
        retcode = self.job.aggregator.end_test(retcode)
        logger.info("RC after generator plugin and aggregator finish: %s", retcode)

        for plugin in self.job.monitoring_plugins:
            logger.info('Stopping monitoring plugin %s', plugin)
            retcode = plugin.end_test(retcode) or retcode
            logger.info('RC after monitoring plugin finish: %s', retcode)

        for plugin in [p for p in self.plugins.values() if
                       p is not self.job.generator_plugin and p not in self.job.monitoring_plugins]:
            logger.info("Stopping %s plugin", plugin)
            try:
                retcode = plugin.end_test(retcode)
            except Exception:  # FIXME too broad exception clause
                logger.exception("Failed stopping plugin %s", plugin)
                if not retcode:
                    retcode = 1
            logger.info("RC after %s plugin finish: %s", plugin, retcode)
        return retcode

    def plugins_post_process(self, retcode):
        """
        Call post_process() on all plugins
        """
        logger.info("Post-processing test with received RC: %s", retcode)
        self.publish("core", "stage", "post_process")

        for plugin in self.plugins.values():
            logger.info("Post-process %s plugin", plugin)
            try:
                retcode = plugin.post_process(retcode)
                logger.info("RC after %s plugin post-processing: %s", plugin, retcode)
            except Exception:  # FIXME too broad exception clause
                logger.exception("Failed post-processing plugin %s", plugin)
                if not retcode:
                    retcode = 1
        return retcode

    def publish_monitoring_data(self, data):
        """sends pending data set to listeners"""
        for plugin in self.monitoring_data_listeners:
            # deep copy to ensure each listener gets it's own copy
            try:
                plugin.monitoring_data(copy.deepcopy(data))
            except Exception:
                logger.exception("Plugin %s failed to process monitoring data", plugin)

    def __setup_taskset(self, affinity, pid=None, args=None):
        """ if pid specified: set process w/ pid `pid` CPU affinity to specified `affinity` core(s)
            if args specified: modify list of args for Popen to start w/ taskset w/ affinity `affinity`
        """
        self.taskset_path = self.get_option(self.SECTION, 'taskset_path')

        if args:
            return [self.taskset_path, '-c', affinity] + args

        if pid:
            args = "%s -pc %s %s" % (self.taskset_path, affinity, pid)
            retcode, stdout, stderr = execute(args, shell=True, poll_period=0.1, catch_out=True)
            logger.debug('taskset for pid %s stdout: %s', pid, stdout)
            if retcode == 0:
                logger.info("Enabled taskset for pid %s with affinity %s", str(pid), affinity)
            else:
                logger.debug('Taskset setup failed w/ retcode :%s', retcode)
                raise KeyError(stderr)

    def _collect_artifacts(self, validation_failed=False):
        logger.info("Artifacts dir: %s. Collecting artifacts", self.artifacts_dir)
        for filename, keep in self.artifact_files.items():
            try:
                self.__collect_file(filename, keep)
            except Exception as ex:
                logger.warning("Failed to collect file %s: %s", filename, ex)

    def get_option(self, section, option, default=None):
        return self.config.get_option(section, option, default)

    def set_option(self, section, option, value):
        """
        Set an option in storage
        """
        raise NotImplementedError

    def set_exitcode(self, code):
        self.output['core']['exitcode'] = code

    def get_plugin_of_type(self, plugin_class):
        """
        Retrieve a plugin of desired class, KeyError raised otherwise
        """
        logger.debug("Searching for plugin: %s", plugin_class)
        matches = [plugin for plugin in self.plugins.values() if isinstance(plugin, plugin_class)]
        if matches:
            if len(matches) > 1:
                plugins = [type(p) for p in matches]
                logger.warning(
                    "More then one plugin of type %s found. Found plugins: %s. Using first one.", plugin_class, plugins,
                )
            return matches[0]
        else:
            raise KeyError("Requested plugin type not found: %s", plugin_class)

    def get_plugins_of_type(self, plugin_class):
        """
        Retrieve a list of plugins of desired class, KeyError raised otherwise
        """
        logger.debug("Searching for plugins: %s", plugin_class)
        matches = [plugin for plugin in self.plugins.values() if isinstance(plugin, plugin_class)]
        if matches:
            return matches
        else:
            raise KeyError("Requested plugin type not found: %s" % plugin_class)

    def get_jobno(self, plugin_name='plugin_lunapark'):
        uploader_plugin = self.plugins[plugin_name]
        return uploader_plugin.lp_job.number

    def __collect_file(self, filename, keep_original=False):
        """
        Move or copy single file to artifacts dir
        """
        dest = self.artifacts_dir + '/' + os.path.basename(filename)
        logger.info("Collecting file: %s to %s", filename, dest)
        if not filename or not os.path.exists(filename):
            logger.info("File not found to collect: %s", filename)
            return

        if os.path.exists(dest):
            # FIXME: 3 find a way to store artifacts anyway
            logger.info("File already exists: %s", dest)
            return

        if keep_original:
            shutil.copy(filename, self.artifacts_dir)
        else:
            shutil.move(filename, self.artifacts_dir)

        os.chmod(dest, 0o644)

    def add_artifact_file(self, filename, keep_original=False):
        """
        Add file to be stored as result artifact on post-process phase
        """
        if filename:
            logger.debug("Adding artifact file to collect (keep=%s): %s", keep_original, filename)
            self.artifact_files[filename] = keep_original

    def add_artifact_to_send(self, lp_requisites, content):
        self.artifacts_to_send.append((lp_requisites, content))

    def apply_shorthand_options(self, options, default_section='DEFAULT'):
        for option_str in options:
            key, value = option_str.split('=')
            try:
                section, option = key.split('.')
            except ValueError:
                section = default_section
                option = key
            logger.debug("Override option: %s => [%s] %s=%s", option_str, section, option, value)
            self.set_option(section, option, value)

    def mkstemp(self, suffix, prefix, directory=None):
        """
        Generate temp file name in artifacts base dir
        and close temp file handle
        """
        if not directory:
            directory = self.artifacts_dir
        fd, fname = tempfile.mkstemp(suffix, prefix, directory)
        os.close(fd)
        os.chmod(fname, 0o644)  # FIXME: chmod to parent dir's mode?
        return fname

    def publish(self, publisher, key, value):
        self.info.update([publisher] + key.split('.'), value)

    def close(self):
        """
        Call close() for all plugins
        """
        logger.info("Close allocated resources...")
        for plugin in self.plugins.values():
            logger.debug("Close %s", plugin)
            try:
                plugin.close()
            except Exception:
                logger.exception("Failed closing plugin %s", plugin)

    @property
    def artifacts_dir(self):
        if not self._artifacts_dir:
            new_path = os.path.join(self.artifacts_base_dir, self.test_id)
            if not os.path.isdir(new_path):
                os.makedirs(new_path)
            os.chmod(new_path, 0o755)
            self._artifacts_dir = os.path.abspath(new_path)
        return self._artifacts_dir

    @staticmethod
    def get_user_agent():
        tank_agent = 'YandexTank/{}'.format(VERSION)
        py_info = sys.version_info
        python_agent = 'Python/{}.{}.{}'.format(
            py_info[0], py_info[1], py_info[2])
        os_agent = 'OS/{}'.format(platform.platform())
        return ' '.join((tank_agent, python_agent, os_agent))

    def register_plugin(self, plugin_name, instance):
        if self._plugins is None:
            self._plugins = {}
        if self._plugins.get(plugin_name, None) is not None:
            logger.error('Plugin with name %s has been already registered', plugin_name)
        self._plugins[plugin_name] = instance

    def save_cfg(self, path):
        self.config.dump(path)

    def plugins_cleanup(self):
        for plugin_name, plugin in self.plugins.items():
            logger.info('Cleaning up plugin %s', plugin_name)
            try:
                plugin.cleanup()
            except Exception:
                logger.exception('Error trying to cleanup plugin %s', plugin_name)


class JobsStorage:

    def __init__(self, file_name=None):
        self.storage_file = file_name or os.getenv(JOBS_STORAGE_FILE_ENV, DEFAULT_JOBS_STORAGE_FILE)
        if not os.path.exists(self.storage_file):
            self._create_storage_file()

    def _create_storage_file(self):
        directory, _ = os.path.split(self.storage_file)
        os.makedirs(directory, exist_ok=True)
        with open(self.storage_file, 'x'):
            os.chmod(self.storage_file, 0o766)

    def push_job(self, cloud_job_id, tank_job_id=None):
        with open(self.storage_file, 'a') as f:
            job_data = {CLOUD_KEY: cloud_job_id, TANK_KEY: tank_job_id}
            logger.info("Push job to storage %s: %s", self.storage_file, job_data)
            yaml.dump([job_data], f)

    def get_cloud_job_id(self, tank_job_id):
        try:
            with open(self.storage_file) as f:
                data = yaml.safe_load(f)
            if data is None:
                return
            for job in data:
                if job[TANK_KEY] == tank_job_id:
                    return job[CLOUD_KEY]
        except FileNotFoundError:
            return
        return


class Lock(object):
    PID = 'pid'
    TEST_ID = 'test_id'
    TEST_DIR = 'test_dir'
    LOCK_FILE_WILDCARD = 'lunapark_*.lock'

    def __init__(self, test_id, test_dir, pid=None):
        self.test_id = test_id
        self.test_dir = test_dir
        self.pid = pid if pid is not None else os.getpid()
        self.info = {
            self.PID: self.pid,
            self.TEST_ID: self.test_id,
            self.TEST_DIR: self.test_dir
        }
        self.lock_file = None

    def acquire(self, lock_dir, ignore=False):
        is_locked = self.is_locked(lock_dir)
        if not ignore and is_locked:
            raise LockError("Lock file(s) found\n{}".format(is_locked))
        prefix, suffix = self.LOCK_FILE_WILDCARD.split('*')
        fh, self.lock_file = tempfile.mkstemp(suffix, prefix, lock_dir)
        os.close(fh)
        with open(self.lock_file, 'w') as f:
            yaml.dump(self.info, f)
        os.chmod(self.lock_file, 0o644)
        return self

    def release(self):
        if self.lock_file is not None and os.path.exists(self.lock_file):
            logger.info("Releasing lock: %s", self.lock_file)
            os.remove(self.lock_file)
        else:
            logger.warning('Lock file not found')

    @classmethod
    def load(cls, path):
        with open(path) as f:
            info = yaml.load(f, Loader=yaml.FullLoader)
        pid = info.get(cls.PID)
        test_id = info.get(cls.TEST_ID)
        test_dir = info.get(cls.TEST_DIR)
        lock = Lock(test_id, test_dir, pid)
        lock.lock_file = path
        return lock

    def __str__(self):
        return str(self.info)

    @classmethod
    def is_locked(cls, lock_dir='/var/lock'):
        for filename in os.listdir(lock_dir):
            if fnmatch.fnmatch(filename, cls.LOCK_FILE_WILDCARD):
                full_name = os.path.join(lock_dir, filename)
                logger.info("Lock file is found: %s", full_name)
                try:
                    running_lock = cls.load(full_name)
                    if not running_lock.pid:
                        msg = 'Failed to get {} from lock file {}.'.format(cls.PID, full_name)
                        logger.warning(msg)
                        return msg
                    else:
                        if not pid_exists(int(running_lock.pid)):
                            logger.info("Lock PID %s not exists, ignoring and trying to remove", running_lock.pid)
                            try:
                                os.remove(full_name)
                            except Exception as exc:
                                logger.warning("Failed to delete lock %s: %s", full_name, exc)
                            return False
                        else:
                            return "Another test is running: {}".format(running_lock)
                except Exception:
                    msg = "Failed to load info from lock %s" % full_name
                    logger.warning(msg, exc_info=True)
                    return msg
        return False

    @classmethod
    def running_ids(cls, lock_dir='/var/lock'):
        return [Lock.load(fname).test_id for fname in glob.glob(os.path.join(lock_dir, cls.LOCK_FILE_WILDCARD))]



''' Module that collects remote system information '''
import getpass
import logging

from yandextank.common.util import SecuredShell
from ...common.interfaces import AbstractPlugin
from ..Phantom import Plugin as PhantomPlugin

logger = logging.getLogger(__name__)


class Plugin(AbstractPlugin):
    '''Plugin that collects remote system information'''
    SECTION = "platform"

    @staticmethod
    def get_key():
        return __file__

    def __init__(self, core, cfg, name):
        AbstractPlugin.__init__(self, core, cfg, name)
        self.hosts = []
        self.port = None
        self.logfile = None
        self.default_target = None

        def _echo_wrapper(cmd):
            return 'echo "====Executing: {cmd}"; {cmd}'.format(cmd=cmd)

        cmds = {
            "dpkg": "dpkg -l",
            "uname": "uname -a",
            "ulimit": "ulimit -a",
            "os_identifier": "cat /etc/issue.net",
            "uptime": "uptime",
            "cpuinfo": "cat /proc/cpuinfo",
            "meminfo": "cat /proc/meminfo",
            "free": "free -m",
            "mounts": "cat /proc/mounts",
            "df": "df -h",
            "ifconfig": "ifconfig -a",
            "sysctl": "cat /etc/sysctl.conf",
            "lsmod": "lsmod"
        }
        self.cmd = "%s" % ";\n".join([_echo_wrapper(cmd) for key, cmd in cmds.items()])

    def get_available_options(self):
        return ["hosts", "port", "username", "timeout"]

    def configure(self):
        try:
            hosts = self.get_option("hosts", "").strip()
            if hosts:
                self.hosts = hosts.split(" ")
            self.port = int(self.get_option("port", 22))
            self.username = self.get_option("username", getpass.getuser())
            self.ssh_key_path = self.get_option("ssh_key_path", None)
            self.timeout = int(self.get_option("timeout", 3))
        except BaseException:
            logger.exception('Exception trying to configure Platform plugin')

        self.logfile = self.core.mkstemp(".log", "platform_")
        self.core.add_artifact_file(self.logfile)

    def prepare_test(self):
        try:
            phantom = self.core.get_plugin_of_type(PhantomPlugin)
            info = phantom.get_info()
            if info:
                if info.address and info.address not in self.hosts:
                    logger.debug("Adding platform check of default_target %s", info.address)
                    self.hosts.append(info.address)
        except KeyError as ex:
            logger.debug(ex)
        for host in self.hosts:
            self.ssh = SecuredShell(
                host, self.port, self.username, command_timeout=self.timeout, ssh_key_path=self.ssh_key_path
            )
            try:
                self.ssh.ensure_connection()
                out, errors, err_code = self.ssh.execute(self.cmd)
            except Exception:
                logger.warning(
                    "Failed to check remote system information at %s:%s", host,
                    self.port)
                logger.debug(
                    "Failed to check remote system information at %s:%s",
                    host,
                    self.port,
                    exc_info=True)
            else:
                # logger.info('Remote system `%s` information: %s', host, out)
                with open(self.logfile, 'w') as f:
                    f.write(out)
                if errors:
                    logging.debug("[%s] error: '%s'", host, errors)

    def is_test_finished(self):
        return -1



''' Cummulative Autostops '''
import logging
import math
import re
from collections import deque

import numpy as np
from ...common.util import expand_to_milliseconds, expand_to_seconds

from .criterions import AbstractCriterion

logger = logging.getLogger(__name__)


class WindowCounter(object):
    def __init__(self, window_size):
        self.window_size = window_size
        self.value = 0.0
        self.q = deque()

    def push(self, value):
        self.value += value
        self.q.append(value)
        if len(self.q) > self.window_size:
            self.value -= self.q.popleft()

    def __len__(self):
        return len(self.q)


class TotalFracTimeCriterion(AbstractCriterion):
    '''
    Windowed time criterion

    syntax: total_time(300ms, 70%, 3s)

    Stop test if 70 percent of response times are greater then 300ms in
    a sliding window of 3 seconds.

    Warning: this criterion uses histogram to make its decision, so the
    time boundary is rounded to aggregator bin edge.
    '''

    @staticmethod
    def get_type_string():
        return 'total_time'

    def __init__(self, autostop, param_str):
        AbstractCriterion.__init__(self)
        self.autostop = autostop
        params = param_str.split(',')
        self.rt_limit = expand_to_milliseconds(params[0]) * 1000
        self.fail_ratio_limit = float(params[1][:-1]) / 100.0
        self.window_size = expand_to_seconds(params[2])
        self.fail_counter = WindowCounter(self.window_size)
        self.total_counter = WindowCounter(self.window_size)
        self.total_fail_ratio = 0.0
        self.seconds = deque()
        self.tag = params[3].strip() if len(params) == 4 else None

    def __fail_count(self, data):
        ecdf = np.cumsum(data["overall"]["interval_real"]["hist"]["data"])
        idx = np.searchsorted(data["overall"]["interval_real"]["hist"]["bins"], self.rt_limit)
        if self.tag:
            if data["tagged"].get(self.tag):
                ecdf = np.cumsum(data["tagged"][self.tag]["interval_real"]["hist"]["data"])
                idx = np.searchsorted(data["tagged"][self.tag]["interval_real"]["hist"]["bins"], self.rt_limit)
            else:
                idx = 0

        if idx == 0:
            return ecdf[-1]
        elif idx == len(ecdf):
            return 0
        else:
            return ecdf[-1] - ecdf[idx]

    def notify(self, data, stat):
        total_responses = self.parse_data(data)
        self.seconds.append((data, stat))
        self.fail_counter.push(self.__fail_count(data))
        self.total_counter.push(total_responses)
        self.total_fail_ratio = (
            self.fail_counter.value / self.total_counter.value)
        if self.total_fail_ratio >= self.fail_ratio_limit and len(
                self.fail_counter) >= self.window_size:
            self.cause_second = self.seconds[0]
            logger.debug(self.explain())
            return True
        if len(self.seconds) > self.window_size:
            self.seconds.popleft()
        return False

    def parse_data(self, data):
        # Parse data for specific tag if it's present
        if self.tag:
            if data["tagged"].get(self.tag):
                total_responses = data["tagged"][self.tag]["interval_real"]["len"]
            else:
                total_responses = data["overall"]["interval_real"]["len"]
        # Parse data for overall
        else:
            total_responses = data["overall"]["interval_real"]["len"]
        return total_responses

    def get_rc(self):
        return self.RC_TOTAL_TIME

    def explain(self):
        items = self.get_criterion_parameters()
        explanation = "%(ratio).2f%% responses times higher than %(limit)sms for %(seconds_count)ss " % items
        if self.tag:
            explanation = explanation + " for tag %(tag)s" % items
        return explanation

    def get_criterion_parameters(self):
        parameters = {
            'ratio': self.total_fail_ratio * 100,
            'limit': self.rt_limit / 1000,
            'seconds_count': self.window_size,
            'tag': self.tag
        }
        return parameters

    def widget_explain(self):
        items = self.get_criterion_parameters()
        return "%(ratio).2f%% times >%(limit)sms for %(seconds_count)ss" % items, self.total_fail_ratio


class TotalHTTPCodesCriterion(AbstractCriterion):
    ''' Cummulative HTTP Criterion '''

    @staticmethod
    def get_type_string():
        return 'total_http'

    def __init__(self, autostop, param_str):
        AbstractCriterion.__init__(self)
        self.seconds_count = 0
        params = param_str.split(',')
        self.codes_mask = params[0].lower()
        self.codes_regex = re.compile(self.codes_mask.replace("x", '.'))
        self.autostop = autostop
        self.data = deque()
        self.second_window = deque()

        level_str = params[1].strip()
        if level_str[-1:] == '%':
            self.level = float(level_str[:-1])
            self.is_relative = True
        else:
            self.level = int(level_str)
            self.is_relative = False
        self.seconds_limit = expand_to_seconds(params[2])
        self.tag = params[3].strip() if len(params) == 4 else None

    def notify(self, data, stat):
        matched_responses, total_responses = self.parse_data(data)
        if self.is_relative:
            if total_responses > 0:
                matched_responses = float(matched_responses) / total_responses * 100
            else:
                matched_responses = 1
        logger.debug(
            "HTTP codes matching mask %s: %s/%s", self.codes_mask,
            matched_responses, self.level)
        self.data.append(matched_responses)
        self.second_window.append((data, stat))
        if len(self.data) > self.seconds_limit:
            self.data.popleft()
            self.second_window.popleft()
        queue_len = 1
        if self.is_relative:
            queue_len = len(self.data)
        if (sum(self.data) / queue_len) >= self.level\
                and len(self.data) >= self.seconds_limit:  # yapf:disable
            self.cause_second = self.second_window[0]
            logger.debug(self.explain())
            return True
        return False

    def parse_data(self, data):
        # Parse data for specific tag if it is present
        if self.tag:
            if data["tagged"].get(self.tag):
                matched_responses = self.count_matched_codes(
                    self.codes_regex, data["tagged"][self.tag]["proto_code"]["count"])
                total_responses = data["tagged"][self.tag]["interval_real"]["len"]
            # matched_responses=0 if current tag differs from selected one
            else:
                matched_responses = 0
                total_responses = data["overall"]["interval_real"]["len"]
        # Parse data for overall
        else:
            matched_responses = self.count_matched_codes(
                self.codes_regex, data["overall"]["proto_code"]["count"])
            total_responses = data["overall"]["interval_real"]["len"]
        return matched_responses, total_responses

    def get_rc(self):
        return self.RC_TOTAL_HTTP

    def get_level_str(self):
        ''' format level str '''
        if self.is_relative:
            level_str = str(self.level) + "%"
        else:
            level_str = self.level
        return level_str

    def explain(self):
        items = self.get_criterion_parameters()
        explanation = "%(code)s codes count higher than %(level)s for %(seconds_limit)ss" % items
        if self.tag:
            explanation = explanation + " for tag %(tag)s" % items
        return explanation

    def get_criterion_parameters(self):
        parameters = {
            'code': self.codes_mask,
            'level': self.get_level_str(),
            'seconds_limit': self.seconds_limit,
            'tag': self.tag
        }
        return parameters

    def widget_explain(self):
        items = self.get_criterion_parameters()
        explanation = "HTTP %(code)s>%(level)s for %(seconds_limit)ss" % items
        if self.is_relative:
            return explanation, sum(self.data)
        return explanation, 1.0


class TotalNetCodesCriterion(AbstractCriterion):
    ''' Cummulative Net Criterion '''

    @staticmethod
    def get_type_string():
        return 'total_net'

    def __init__(self, autostop, param_str):
        AbstractCriterion.__init__(self)
        self.seconds_count = 0
        params = param_str.split(',')
        self.codes_mask = params[0].lower()
        self.codes_regex = re.compile(self.codes_mask.replace("x", '.'))
        self.autostop = autostop
        self.data = deque()
        self.second_window = deque()

        level_str = params[1].strip()
        if level_str[-1:] == '%':
            self.level = float(level_str[:-1])
            self.is_relative = True
        else:
            self.level = int(level_str)
            self.is_relative = False
        self.seconds_limit = expand_to_seconds(params[2])
        self.tag = params[3].strip() if len(params) == 4 else None

    def notify(self, data, stat):
        matched_responses, total_responses = self.parse_data(data)
        if self.is_relative:
            if total_responses:
                matched_responses = float(matched_responses) / total_responses * 100
                logger.debug(
                    "Net codes matching mask %s: %s%%/%s", self.codes_mask,
                    round(matched_responses, 2), self.get_level_str())
            else:
                matched_responses = 1
        else:
            logger.debug(
                "Net codes matching mask %s: %s/%s", self.codes_mask,
                matched_responses, self.get_level_str())

        self.data.append(matched_responses)
        self.second_window.append((data, stat))
        if len(self.data) > self.seconds_limit:
            self.data.popleft()
            self.second_window.popleft()

        queue_len = 1
        if self.is_relative:
            queue_len = len(self.data)
        if (sum(self.data) / queue_len) >= self.level\
                and len(self.data) >= self.seconds_limit:  # yapf:disable
            self.cause_second = self.second_window[0]
            logger.debug(self.explain())
            return True
        return False

    def parse_data(self, data):
        # Count data for specific tag if it's present
        if self.tag:
            if data["tagged"].get(self.tag):
                codes = data["tagged"][self.tag]["net_code"]["count"].copy()
                if '0' in codes:
                    codes.pop('0')
                matched_responses = self.count_matched_codes(self.codes_regex, codes)
                total_responses = data["tagged"][self.tag]["interval_real"]["len"]
            # matched_responses=0 if current tag differs from selected one
            else:
                matched_responses = 0
                total_responses = data["overall"]["interval_real"]["len"]
        # Count data for overall
        else:
            codes = data["overall"]["net_code"]["count"].copy()
            if '0' in codes:
                codes.pop('0')
            matched_responses = self.count_matched_codes(self.codes_regex, codes)
            total_responses = data["overall"]["interval_real"]["len"]
        return matched_responses, total_responses

    def get_rc(self):
        return self.RC_TOTAL_NET

    def get_level_str(self):
        ''' format level str '''
        if self.is_relative:
            level_str = str(self.level) + "%"
        else:
            level_str = str(self.level)
        return level_str

    def explain(self):
        items = self.get_criterion_parameters()
        explanation = "%(code)s net codes count higher than %(level)s for %(seconds_limit)ss" % items
        if self.tag:
            explanation = explanation + " for tag %(tag)s" % items
        return explanation

    def get_criterion_parameters(self):
        parameters = {
            'code': self.codes_mask,
            'level': self.get_level_str(),
            'seconds_limit': self.seconds_limit,
            'tag': self.tag
        }
        return parameters

    def widget_explain(self):
        items = self.get_criterion_parameters()
        explanation = "Net %(code)s>%(level)s for %(seconds_limit)ss" % items
        if self.is_relative:
            return explanation, sum(self.data)
        return explanation, 1.0


class TotalNegativeHTTPCodesCriterion(AbstractCriterion):
    ''' Reversed HTTP Criterion '''

    @staticmethod
    def get_type_string():
        return 'negative_http'

    def __init__(self, autostop, param_str):
        AbstractCriterion.__init__(self)
        self.seconds_count = 0
        params = param_str.split(',')
        self.codes_mask = params[0].lower()
        self.codes_regex = re.compile(self.codes_mask.replace("x", '.'))
        self.autostop = autostop
        self.data = deque()
        self.second_window = deque()

        level_str = params[1].strip()
        if level_str[-1:] == '%':
            self.level = float(level_str[:-1])
            self.is_relative = True
        else:
            self.level = int(level_str)
            self.is_relative = False
        self.seconds_limit = expand_to_seconds(params[2])
        self.tag = params[3].strip() if len(params) == 4 else None

    def notify(self, data, stat):
        matched_responses, total_responses = self.parse_data(data)
        if self.is_relative:
            if total_responses:
                matched_responses = float(matched_responses) / total_responses * 100
                matched_responses = 100 - matched_responses
            else:
                matched_responses = 1
            logger.debug(
                "HTTP codes matching mask not %s: %s/%s", self.codes_mask,
                round(matched_responses, 1), self.level)
        else:
            matched_responses = (
                total_responses - matched_responses)
            logger.debug(
                "HTTP codes matching mask not %s: %s/%s", self.codes_mask,
                matched_responses, self.level)
        self.data.append(matched_responses)
        self.second_window.append((data, stat))
        if len(self.data) > self.seconds_limit:
            self.data.popleft()
            self.second_window.popleft()

        queue_len = 1
        if self.is_relative:
            queue_len = len(self.data)
        if (sum(self.data) / queue_len) >= self.level\
                and len(self.data) >= self.seconds_limit:  # yapf:disable
            self.cause_second = self.second_window[0]
            logger.debug(self.explain())
            return True
        return False

    def parse_data(self, data):
        # Parse data for specific tag if it is present
        if self.tag:
            if data["tagged"].get(self.tag):
                matched_responses = self.count_matched_codes(
                    self.codes_regex, data["tagged"][self.tag]["proto_code"]["count"])
                total_responses = data["tagged"][self.tag]["interval_real"]["len"]
            # matched_responses=0 if current tag differs from selected one
            else:
                matched_responses = 0
                total_responses = data["overall"]["interval_real"]["len"]
        # Parse data for overall
        else:
            matched_responses = self.count_matched_codes(
                self.codes_regex, data["overall"]["proto_code"]["count"])
            total_responses = data["overall"]["interval_real"]["len"]
        return matched_responses, total_responses

    def get_rc(self):
        return self.RC_TOTAL_NEGATIVE_HTTP

    def get_level_str(self):
        ''' format level str'''
        if self.is_relative:
            level_str = str(self.level) + "%"
        else:
            level_str = self.level
        return level_str

    def explain(self):
        items = self.get_criterion_parameters()
        explanation = "Not %(code)s codes count higher than %(level)s for %(seconds_limit)ss" % items
        if self.tag:
            explanation = explanation + " for tag %(tag)s" % items
        return explanation

    def get_criterion_parameters(self):
        parameters = {
            'code': self.codes_mask,
            'level': self.get_level_str(),
            'seconds_limit': self.seconds_limit,
            'tag': self.tag
        }
        return parameters

    def widget_explain(self):
        items = self.get_criterion_parameters()
        explanation = "HTTP not %(code)s>%(level)s for %(seconds_limit)ss" % items
        if self.is_relative:
            return explanation, sum(self.data)
        return explanation, 1.0


class TotalNegativeNetCodesCriterion(AbstractCriterion):
    ''' Reversed NET Criterion '''

    @staticmethod
    def get_type_string():
        return 'negative_net'

    def __init__(self, autostop, param_str):
        AbstractCriterion.__init__(self)
        self.seconds_count = 0
        params = param_str.split(',')
        self.codes_mask = params[0].lower()
        self.codes_regex = re.compile(self.codes_mask.replace("x", '.'))
        self.autostop = autostop
        self.data = deque()
        self.second_window = deque()

        level_str = params[1].strip()
        if level_str[-1:] == '%':
            self.level = float(level_str[:-1])
            self.is_relative = True
        else:
            self.level = int(level_str)
            self.is_relative = False
        self.seconds_limit = expand_to_seconds(params[2])
        self.tag = params[3].strip() if len(params) == 4 else None

    def notify(self, data, stat):
        matched_responses, total_responses = self.parse_data(data)
        if self.is_relative:
            if total_responses:
                matched_responses = float(matched_responses) / total_responses * 100
                matched_responses = 100 - matched_responses
            else:
                matched_responses = 1
            logger.debug(
                "Net codes matching mask not %s: %s/%s", self.codes_mask,
                round(matched_responses, 1), self.level)
        else:
            matched_responses = (
                total_responses - matched_responses)
            logger.debug(
                "Net codes matching mask not %s: %s/%s", self.codes_mask,
                matched_responses, self.level)
        self.data.append(matched_responses)
        self.second_window.append((data, stat))
        if len(self.data) > self.seconds_limit:
            self.data.popleft()
            self.second_window.popleft()

        queue_len = 1
        if self.is_relative:
            queue_len = len(self.data)
        if (sum(self.data) / queue_len) >= self.level \
                and len(self.data) >= self.seconds_limit:  # yapf:disable
            self.cause_second = self.second_window[0]
            logger.debug(self.explain())
            return True
        return False

    def parse_data(self, data):
        # Count data for specific tag if it's present
        if self.tag:
            if data["tagged"].get(self.tag):
                matched_responses = self.count_matched_codes(self.codes_regex,
                                                             data["tagged"][self.tag]["net_code"]["count"])
                total_responses = data["tagged"][self.tag]["interval_real"]["len"]
            # matched_responses=0 if current tag differs from selected one
            else:
                matched_responses = 0
                total_responses = data["overall"]["interval_real"]["len"]
        # Count data for overall
        else:
            matched_responses = self.count_matched_codes(self.codes_regex,
                                                         data["overall"]["net_code"]["count"])
            total_responses = data["overall"]["interval_real"]["len"]
        return matched_responses, total_responses

    def get_rc(self):
        return self.RC_TOTAL_NEGATIVE_NET

    def get_level_str(self):
        ''' format level str'''
        if self.is_relative:
            level_str = str(self.level) + "%"
        else:
            level_str = self.level
        return level_str

    def explain(self):
        items = self.get_criterion_parameters()
        explanation = "Not %(code)s codes count higher than %(level)s for %(seconds_limit)ss" % items
        if self.tag:
            explanation = explanation + " for tag %(tag)s" % items
        return explanation

    def get_criterion_parameters(self):
        parameters = {
            'code': self.codes_mask,
            'level': self.get_level_str(),
            'seconds_limit': self.seconds_limit,
            'tag': self.tag
        }
        return parameters

    def widget_explain(self):
        items = self.get_criterion_parameters()
        explanation = "Net not %(code)s>%(level)s for %(seconds_limit)ss" % items
        if self.is_relative:
            return explanation, sum(self.data)
        return explanation, 1.0


class TotalHTTPTrendCriterion(AbstractCriterion):
    ''' HTTP Trend Criterion '''

    @staticmethod
    def get_type_string():
        return 'http_trend'

    def __init__(self, autostop, param_str):
        AbstractCriterion.__init__(self)
        self.seconds_count = 0
        params = param_str.split(',')
        self.codes_mask = params[0].lower()
        self.codes_regex = re.compile(self.codes_mask.replace("x", '.'))
        self.autostop = autostop
        self.tangents = deque()
        self.second_window = deque()
        self.total_tan = float()

        self.tangents.append(0)
        self.last = 0
        self.seconds_limit = expand_to_seconds(params[1])
        self.measurement_error = float()
        self.tag = params[2].strip() if len(params) == 3 else None

    def notify(self, data, stat):
        matched_responses = self.parse_data(data)
        self.tangents.append(matched_responses - self.last)
        self.second_window.append((data, stat))

        self.last = matched_responses

        if len(self.tangents) > self.seconds_limit:
            self.tangents.popleft()
            self.second_window.popleft()

        self.measurement_error = self.calc_measurement_error(self.tangents)

        self.total_tan = float(sum(self.tangents) / len(self.tangents))
        logger.debug(
            "Last trend for http codes %s: %.2f +/- %.2f", self.codes_mask,
            self.total_tan, self.measurement_error)

        if self.total_tan + self.measurement_error < 0:
            self.cause_second = self.second_window[0]
            logger.debug(self.explain())
            return True

        return False

    def parse_data(self, data):
        # Count data for specific tag if it's present
        if self.tag:
            if data["tagged"].get(self.tag):
                matched_responses = self.count_matched_codes(
                    self.codes_regex, data["tagged"][self.tag]["proto_code"]["count"])
            # matched_responses=0 if current tag differs from selected one
            else:
                matched_responses = 0
        # Count data for overall if it's present
        else:
            matched_responses = self.count_matched_codes(
                self.codes_regex, data["overall"]["proto_code"]["count"])
        return matched_responses

    def calc_measurement_error(self, tangents):
        '''
        formula for measurement error
        sqrt ( (sum(1, n, (k_i - <k>)**2) / (n*(n-1)))
        '''

        if len(tangents) < 2:
            return 0.0

        avg_tan = float(sum(tangents) / len(tangents))
        numerator = float()
        for i in tangents:
            numerator += (i - avg_tan) * (i - avg_tan)

        return math.sqrt(numerator / len(tangents) / (len(tangents) - 1))

    def get_rc(self):
        return self.RC_TOTAL_HTTP_TREND

    def explain(self):
        items = self.get_criterion_parameters()
        return "Last trend for %(code)s http codes " \
            "is %(total_tan).2f +/- %(measurement_err).2f for %(seconds_limit)ss" % items

    def get_criterion_parameters(self):
        parameters = {
            'code': self.codes_mask,
            'total_tan': self.total_tan,
            'measurement_err': self.measurement_error,
            'seconds_limit': self.seconds_limit,
            'tag': self.tag
        }
        return parameters

    def widget_explain(self):
        items = self.get_criterion_parameters()
        return "HTTP(%(code)s) trend is %(total_tan).2f +/- %(measurement_err).2f < 0 for %(seconds_limit)ss" % items, 1.0



""" Utility classes for phantom module """
# TODO: use separate answ log per benchmark
import copy
import logging
import multiprocessing
import string

from pkg_resources import resource_string
from yandextank.common.util import AddressWizard

from ...stepper import StepperWrapper
from ...stepper.util import parse_duration

logger = logging.getLogger(__name__)


class PhantomConfig:
    """ config file generator """
    OPTION_PHOUT = "phout_file"

    def __init__(self, core, cfg, stat_log):
        self.core = core
        self.cfg = cfg
        self.streams = []

        # common
        self.timeout = 11000
        self.answ_log = None
        self.answ_log_level = None
        self._phout_file = None
        self.stat_log = stat_log
        self.phantom_log = None
        self.phantom_start_time = None
        self.phantom_modules_path = None
        self.threads = None
        self.additional_libs = None
        self.enum_ammo = False
        self._config_file = None

    @staticmethod
    def get_available_options():
        opts = [
            "threads",
            "phantom_modules_path",
            "additional_libs",
            "writelog",
            "enum_ammo",
            "timeout",
        ]
        opts += StreamConfig.get_available_options()
        return opts

    @property
    def phout_file(self):
        if not self._phout_file:
            self._phout_file = self.cfg['phout_file'] or self.core.mkstemp(".log", "phout_")
        return self._phout_file

    def read_config(self):
        """        Read phantom tool specific options        """
        self.threads = self.cfg["threads"] or str(min(int(multiprocessing.cpu_count() / 2) + 1, 128))
        self.phantom_modules_path = self.cfg["phantom_modules_path"]
        self.additional_libs = ' '.join(self.cfg["additional_libs"])
        self.answ_log_level = self.cfg["writelog"]
        if self.answ_log_level.lower() in ['0', 'false']:
            self.answ_log_level = 'none'
        elif self.answ_log_level.lower() in ['1', 'true']:
            self.answ_log_level = 'all'
        self.timeout = parse_duration(self.cfg["timeout"])
        if self.timeout > 120000:
            logger.warning(
                "You've set timeout over 2 minutes."
                " Are you a functional tester?")
        self.answ_log = self.core.mkstemp(".log", "answ_")
        self.core.add_artifact_file(self.answ_log)
        self.core.add_artifact_file(self.phout_file)
        self.core.add_artifact_file(self.stat_log)
        self.phantom_log = self.core.mkstemp(".log", "phantom_")
        self.core.add_artifact_file(self.phantom_log)

        main_stream = StreamConfig(
            self.core,
            len(self.streams), self.phout_file, self.answ_log,
            self.answ_log_level, self.timeout, self.cfg, True)
        self.streams.append(main_stream)

        for section in self.multi():
            self.streams.append(
                StreamConfig(
                    self.core,
                    len(self.streams), self.phout_file, self.answ_log,
                    self.answ_log_level, self.timeout, section))

        for stream in self.streams:
            stream.read_config()

        if any(stream.ssl for stream in self.streams):
            self.additional_libs += ' ssl io_benchmark_method_stream_transport_ssl'

    def multi(self):
        return (dict(self.cfg, **section) for section in self.cfg['multi'])

    @property
    def config_file(self):
        if self._config_file is None:
            self._config_file = self.compose_config()
        return self._config_file

    def compose_config(self):
        """        Generate phantom tool run config        """
        streams_config = ''
        stat_benchmarks = ''
        for stream in self.streams:
            streams_config += stream.compose_config()
            if not stream.is_main:
                stat_benchmarks += " " + "benchmark_io%s" % stream.sequence_no

        kwargs = {}
        kwargs['threads'] = self.threads
        kwargs['phantom_log'] = self.phantom_log
        kwargs['stat_log'] = self.stat_log
        kwargs['benchmarks_block'] = streams_config
        kwargs['stat_benchmarks'] = stat_benchmarks
        kwargs['additional_libs'] = self.additional_libs
        kwargs['phantom_modules_path'] = self.phantom_modules_path
        filename = self.core.mkstemp(".conf", "phantom_")
        self.core.add_artifact_file(filename)
        logger.debug("Generating phantom config: %s", filename)
        template_str = resource_string(__name__, "config/phantom.conf.tpl").decode('utf8')
        tpl = string.Template(template_str)
        config = tpl.substitute(kwargs)

        with open(filename, 'w') as conffile:
            conffile.write(config)
        return filename

    def set_timeout(self, timeout):
        """ pass timeout to all streams """
        for stream in self.streams:
            stream.timeout = timeout

    def get_info(self):
        """ get merged info about phantom conf """
        result = copy.copy(self.streams[0])
        result.stat_log = self.stat_log
        result.steps = []
        result.ammo_file = ''
        result.rps_schedule = None
        result.ammo_count = 0
        result.duration = 0

        result.instances = 0
        result.loadscheme = []
        result.loop_count = 0

        for stream in self.streams:
            sec_no = 0
            logger.debug("Steps: %s", stream.stepper_wrapper.steps)
            for item in stream.stepper_wrapper.steps:
                for x in range(0, int(item[1])):
                    if len(result.steps) > sec_no:
                        result.steps[sec_no][0] += item[0]
                    else:
                        result.steps.append([item[0], 1])
                    sec_no += 1

            if result.rps_schedule:
                result.rps_schedule = []
            else:
                result.rps_schedule = stream.stepper_wrapper.loadscheme
            if result.loadscheme:
                result.loadscheme = ''
            else:
                # FIXME: add formatted load scheme for server:
                # <step_size,step_type,first_rps,last_rps,original_step_params>
                # as a string
                result.loadscheme = ''

            if result.loop_count:
                result.loop_count = '0'
            else:
                result.loop_count = stream.stepper_wrapper.loop_count

            result.ammo_file += '{} '.format(stream.stepper_wrapper.ammo_file)
            result.ammo_count += stream.stepper_wrapper.ammo_count
            result.duration = max(
                result.duration, stream.stepper_wrapper.duration)
            result.instances += stream.instances

        if not result.ammo_count:
            raise ValueError("Total ammo count cannot be zero")
        return result


class StreamConfig:
    """ each test stream's config """

    OPTION_INSTANCES_LIMIT = 'instances'

    def __init__(self, core, sequence, phout, answ, answ_level, timeout, cfg, is_main=False):
        self.core = core
        self.cfg = cfg
        self.address_wizard = AddressWizard()

        self.sequence_no = sequence
        self.stepper_wrapper = StepperWrapper(core, cfg)
        self.phout_file = phout
        self.answ_log = answ
        self.answ_log_level = answ_level
        self.timeout = timeout
        self.is_main = is_main

        # per benchmark
        self.instances = self.get_option('instances')
        self.ipv6 = None
        self.ssl = None
        self.address = None
        self.port = None
        self.tank_type = None
        self.stpd = None
        self.gatling = None
        self.phantom_http_line = None
        self.phantom_http_field_num = None
        self.phantom_http_field = None
        self.phantom_http_entity = None
        self.resolved_ip = None
        self.method_prefix = None
        self.source_log_prefix = None
        self.method_options = None
        self.client_cipher_suites = None
        self.client_certificate = None
        self.client_key = None

    def get_option(self, option, default=None):
        """ get option wrapper """
        return self.cfg[option]

    @staticmethod
    def get_available_options():
        opts = [
            "ssl", "tank_type", 'gatling_ip', "method_prefix",
            "source_log_prefix"
        ]
        opts += [
            "phantom_http_line", "phantom_http_field_num", "phantom_http_field",
            "phantom_http_entity"
        ]
        opts += ['address', "port", StreamConfig.OPTION_INSTANCES_LIMIT]
        opts += StepperWrapper.get_available_options()
        opts += ["connection_test"]
        return opts

    def read_config(self):
        """ reads config """
        # multi-options
        self.ssl = self.get_option("ssl")
        self.tank_type = self.get_option("tank_type")
        # TODO: refactor. Maybe we should decide how to interact with
        # StepperWrapper here.
        # self.instances = self.get_option('instances')
        self.gatling = ' '.join(self.get_option('gatling_ip').split("\n"))
        self.method_prefix = self.get_option("method_prefix")
        self.method_options = self.get_option("method_options")
        self.source_log_prefix = self.get_option("source_log_prefix")

        self.phantom_http_line = self.get_option("phantom_http_line")
        self.phantom_http_field_num = self.get_option("phantom_http_field_num")
        self.phantom_http_field = self.get_option("phantom_http_field")
        self.phantom_http_entity = self.get_option("phantom_http_entity")

        self.address = self.get_option('address')
        do_test_connect = self.get_option("connection_test")
        explicit_port = self.get_option('port', '')
        self.ipv6, self.resolved_ip, self.port, self.address = self.address_wizard.resolve(
            self.address, do_test_connect, explicit_port)

        logger.info("Resolved %s into %s:%s", self.address, self.resolved_ip, self.port)

        self.client_cipher_suites = self.get_option("client_cipher_suites", "")
        self.client_certificate = self.get_option("client_certificate", "")
        self.client_key = self.get_option("client_key", "")
        self.stepper_wrapper.read_config()

    def compose_config(self):
        """ compose benchmark block """
        # step file
        self.stepper_wrapper.prepare_stepper()
        self.stpd = self.stepper_wrapper.stpd
        if self.stepper_wrapper.instances:
            self.instances = self.stepper_wrapper.instances

        if not self.stpd:
            raise RuntimeError("Cannot proceed with no STPD file")

        kwargs = {}
        kwargs['sequence_no'] = self.sequence_no
        if self.ssl:
            _auth_section = ''
            _ciphers = ''
            ssl_template = "transport_t ssl_transport = transport_ssl_t {\n" \
                           "                timeout = 1s\n" \
                           "                %s\n" \
                           "                %s}\n" \
                           "                transport = ssl_transport"

            if self.client_certificate or self.client_key:
                _auth_section = 'auth_t def_auth = auth_t { key = "%s" cert = "%s"} auth = def_auth' \
                                % (self.client_key, self.client_certificate)
            if self.client_cipher_suites:
                _ciphers = 'ciphers = "%s"' % self.client_cipher_suites
            kwargs['ssl_transport'] = ssl_template % (_auth_section, _ciphers)
        else:
            kwargs['ssl_transport'] = ""
        kwargs['method_stream'] = self.method_prefix + \
            "_ipv6_t" if self.ipv6 else self.method_prefix + "_ipv4_t"
        kwargs['phout'] = self.phout_file
        kwargs['answ_log'] = self.answ_log
        kwargs['answ_log_level'] = self.answ_log_level
        kwargs['comment_answ'] = "# " if self.answ_log_level == 'none' else ''
        kwargs['stpd'] = self.stpd
        kwargs['source_log_prefix'] = self.source_log_prefix
        kwargs['method_options'] = self.method_options
        if self.tank_type:
            kwargs[
                'proto'] = "proto=http_proto%s" % self.sequence_no if self.tank_type == 'http' else "proto=none_proto"
            kwargs['comment_proto'] = ""
        else:
            kwargs['proto'] = ""
            kwargs['comment_proto'] = "#"

        if self.gatling:
            kwargs['bind'] = 'bind={ ' + self.gatling + ' }'
        else:
            kwargs['bind'] = ''
        kwargs['ip'] = self.resolved_ip
        kwargs['port'] = self.port
        kwargs['timeout'] = self.timeout
        kwargs['instances'] = self.instances
        tune = ''
        if self.phantom_http_entity:
            tune += "entity = " + self.phantom_http_entity + "\n"
        if self.phantom_http_field:
            tune += "field = " + self.phantom_http_field + "\n"
        if self.phantom_http_field_num:
            tune += "field_num = {}\n".format(self.phantom_http_field_num)
        if self.phantom_http_line:
            tune += "line = " + self.phantom_http_line + "\n"
        if tune:
            kwargs['reply_limits'] = 'reply_limits = {\n' + tune + "}"
        else:
            kwargs['reply_limits'] = ''

        if self.is_main:
            fname = 'phantom_benchmark_main.tpl'
        else:
            fname = 'phantom_benchmark_additional.tpl'
        template_str = resource_string(
            __name__, "config/" + fname).decode('utf8')
        tpl = string.Template(template_str)
        config = tpl.substitute(kwargs)

        return config


# ========================================================================



#!/usr/bin/env python3
""" The agent bundle, contains all metric classes and agent running code """
import subprocess
import logging
import os
import sys
import signal
import configparser
import json
import threading
import time

from argparse import ArgumentParser
import queue as q

logger = logging.getLogger("agent")
collector_logger = logging.getLogger("telegraf")


def signal_handler(sig, frame):
    """ required for non-tty python runs to interrupt """
    logger.warning("Got signal %s, going to stop", sig)
    raise KeyboardInterrupt()


def ignore_handler(sig, frame):
    logger.warning("Got signal %s, ignoring", sig)


def set_sig_handler():
    uncatchable = ['SIG_DFL', 'SIGSTOP', 'SIGKILL']
    ignore = ['SIGCHLD', 'SIGCLD']
    all_sig = [s for s in dir(signal) if s.startswith("SIG")]
    for sig_name in ignore:
        try:
            sig_num = getattr(signal, sig_name)
            signal.signal(sig_num, ignore_handler)
        except Exception:
            pass
    for sig_name in [s for s in all_sig if s not in (uncatchable + ignore)]:
        try:
            sig_num = getattr(signal, sig_name)
            signal.signal(sig_num, signal_handler)
        except Exception as ex:
            logger.error("Can't set handler for %s, %s", sig_name, ex)


class DataReader(object):
    """generator reads from source line-by-line"""

    def __init__(self, filename, pipe=False):
        self.buffer = ""
        self.closed = False
        self.broken = False
        self.pipe = pipe

        if not self.pipe:
            try:
                self.monout = open(filename, 'rb')
            except Exception:
                logger.error("Can't open source file %s: %s", filename, exc_info=True)
                self.broken = True
        else:
            self.monout = filename

    def __iter__(self):
        while not self.closed:
            if self.broken:
                data = ''
            else:
                data = self.monout.readline().decode('utf8')
            if data:
                parts = data.rsplit('\n', 1)
                if len(parts) > 1:
                    ready_chunk = self.buffer + parts[0] + '\n'
                    self.buffer = parts[1]
                    yield ready_chunk
                else:
                    self.buffer += parts[0]
            else:
                yield None
        if not self.pipe:
            self.monout.close()

    def close(self):
        self.closed = True


class Consolidator(object):
    """generator consolidates data from source, cache it by timestamp"""

    def __init__(self, sources):
        self.sources = sources
        self.results = {}

    def append_chunk(self, source, chunk):
        try:
            data = json.loads(chunk)
        except ValueError:
            logger.error('unable to decode chunk %s', chunk, exc_info=True)
        else:
            try:
                ts = data['timestamp']
                self.results.setdefault(ts, {})
                for key, value in data['fields'].items():
                    if data['name'] == 'diskio':
                        data['name'] = "{metric_name}-{disk_id}".format(
                            metric_name=data['name'],
                            disk_id=data['tags']['name'])
                    elif data['name'] == 'net':
                        data['name'] = "{metric_name}-{interface}".format(
                            metric_name=data['name'],
                            interface=data['tags']['interface'])
                    elif data['name'] == 'cpu':
                        data['name'] = "{metric_name}-{cpu_id}".format(
                            metric_name=data['name'],
                            cpu_id=data['tags']['cpu'])
                    key = data['name'] + "_" + key
                    if key.endswith('_exec_value'):
                        key = key.replace('_exec_value', '')
                    self.results[ts][key] = value
            except KeyError:
                logger.error(
                    'Malformed json from source %s: %s',
                    source,
                    chunk,
                    exc_info=True)
            except BaseException:
                logger.exception('Something nasty happend in consolidator work')

    def __iter__(self):
        while True:
            for s in self.sources:
                chunk_limit = 10
                chunks_done = 0
                chunk = next(s)
                while chunk and chunks_done < chunk_limit:
                    self.append_chunk(s, chunk)
                    chunk = next(s)
                if len(self.results) > 2:
                    logger.debug('Now in buffer: %s', list(self.results.keys()))
                    dump_seconds = sorted(list(self.results.keys()))[:-2]
                    for ready_second in dump_seconds:
                        yield json.dumps({
                            ready_second: self.results.pop(ready_second, None)
                        })
                time.sleep(0.5)


class Drain(threading.Thread):
    """
    Drain a generator to a destination that answers to put(), in a thread
    """

    def __init__(self, source, destination):
        super(Drain, self).__init__()
        self.source = source
        self.destination = destination
        self._finished = threading.Event()
        self._interrupted = threading.Event()

    def run(self):
        for item in self.source:
            self.destination.put(item)
            if self._interrupted.is_set():
                break
        self._finished.set()

    def wait(self, timeout=None):
        self._finished.wait(timeout=timeout)

    def close(self):
        self._interrupted.set()


class AgentWorker(threading.Thread):
    def __init__(self, telegraf_path):
        super(AgentWorker, self).__init__()
        self.working_dir = os.path.dirname(__file__)
        self.startups = []
        self.startup_processes = []
        self.shutdowns = []
        self.custom_sources = []
        self.daemon = True  # Thread auto-shutdown
        self.finished = False
        self.drain = None
        self.drain_stdout = None
        self.drain_err = None
        self.data_reader = None
        self.telegraf_path = telegraf_path
        self.results = q.Queue()
        self.results_stdout = q.Queue()
        self.results_err = q.Queue()

    @staticmethod
    def __popen(cmnd, shell=False):
        logger.info('Starting telegraf binary:\n{}'.format(' '.join(cmnd)))
        return subprocess.Popen(
            cmnd,
            bufsize=0,
            preexec_fn=os.setsid,
            close_fds=True,
            shell=shell,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            stdin=subprocess.PIPE, )

    def read_startup_config(self, cfg_file='agent_startup.cfg'):
        try:
            config = configparser.RawConfigParser(strict=False)
            with open(os.path.join(self.working_dir, cfg_file), 'r') as f:
                config.read_file(f)

            if config.has_section('startup'):
                for option in config.options('startup'):
                    if option.startswith('cmd'):
                        self.startups.append(config.get('startup', option))

            if config.has_section('shutdown'):
                for option in config.options('shutdown'):
                    if option.startswith('cmd'):
                        self.shutdowns.append(config.get('shutdown', option))

            if config.has_section('source'):
                for option in config.options('source'):
                    if option.startswith('file'):
                        self.custom_sources.append(
                            config.get('source', option))
            logger.info(
                'Successfully loaded startup config.\n'
                'Startups: %s\n'
                'Shutdowns: %s\n', self.startups, self.shutdowns)
        except BaseException:
            logger.exception('Error trying to read agent startup config')

    def run(self):
        logger.info("Running startup commands")
        for cmnd in self.startups:
            logger.debug("Run: %s", cmnd)
            # fixme: shell=True is insecure, should save startup script and
            # launch directly
            proc = self.__popen(cmnd, shell=True)
            logger.info('Started with pid %d', proc.pid)
            self.startup_processes.append(proc)

        logger.info('Starting metrics collector..')
        # todo: add identificators into {} for python 2.6
        args = [self.telegraf_path, '-config',
                '{0}/agent.cfg'.format(self.working_dir)]
        self.collector = self.__popen(cmnd=args)
        logger.info('Started with pid %d', self.collector.pid)

        telegraf_output = self.working_dir + '/monitoring.rawdata'
        sources = [telegraf_output] + self.custom_sources

        for _ in range(10):
            self.collector.poll()
            if not self.collector.returncode:
                logger.info("Waiting for telegraf...")
            else:
                logger.info("Telegraf with pid %d ended with code %d", self.collector.pid, self.collector.returncode)
            if os.path.isfile(telegraf_output):
                break
            time.sleep(1)

        self.drain = Drain(
            Consolidator([iter(DataReader(f)) for f in sources]), self.results)
        self.drain.start()

        self.drain_stdout = Drain(
            DataReader(
                self.collector.stdout, pipe=True), self.results_stdout)
        self.drain_stdout.start()

        self.drain_err = Drain(
            DataReader(
                self.collector.stderr, pipe=True), self.results_err)
        self.drain_err.start()

        while not self.finished:
            for _ in range(self.results.qsize()):
                try:
                    data = self.results.get_nowait()
                    logger.debug('send %s bytes of data to collector', len(data))
                    sys.stdout.write(str(data) + '\n')
                    sys.stdout.flush()
                except q.Empty:
                    break
                except BaseException:
                    logger.exception('Something nasty happend trying to send data')
            for _ in range(self.results_stdout.qsize()):
                try:
                    data = self.results_stdout.get_nowait()
                    if data:
                        collector_logger.info("STDOUT: %s", data)
                except q.Empty:
                    break
            for _ in range(self.results_err.qsize()):
                try:
                    data = self.results_err.get_nowait()
                    if data:
                        collector_logger.info("STDERR: %s", data.rstrip('\n'))
                except q.Empty:
                    break
            time.sleep(1)

        self.drain.close()
        self.drain_stdout.close()
        self.drain_err.close()
        self.stop()

    @staticmethod
    def proc_stop(proc, kill=False):
        proc.poll()
        if proc.returncode is None:
            try:
                if kill:
                    logger.info("Killing PID %s", proc.pid)
                    os.killpg(proc.pid, signal.SIGKILL)
                else:
                    logger.info("Terminating: %s", proc.pid)
                    os.killpg(proc.pid, signal.SIGTERM)
                    proc.wait()
                    logger.info('Retcode for PID %s %s', proc.pid, proc.returncode)
            except OSError as ex:
                if ex.errno == 3:
                    logger.info("PID %s already died", proc.pid)

    def kill(self):
        logger.info("Forced stop")
        for proc in self.startup_processes:
            self.proc_stop(proc, kill=True)
        self.proc_stop(self.collector, kill=True)

    def stop(self):
        logger.info("Terminating startup commands")
        for proc in self.startup_processes:
            self.proc_stop(proc)

        logger.info('Terminating collector process: %s', self.collector)
        self.proc_stop(self.collector)

        logger.info("Running shutdown commands")
        for cmnd in self.shutdowns:
            logger.debug("Run: %s", cmnd)
            subprocess.call(cmnd, shell=True)

        self.finished = True
        logger.info("Worker thread finished")
        sys.stderr.write('stopped\n')


def kill_old_agents(telegraf_path):
    my_pid = os.getpid()
    parent = os.getppid()
    logger.info('My pid: {0} Parent pid: {1}'.format(my_pid, parent))
    ps_output = subprocess.check_output(['ps', 'aux'])
    for line in ps_output.splitlines():
        if telegraf_path in line:
            pid = int(line.split()[1])
            logger.info('Found pid: {0}'.format(pid))
            if pid not in [my_pid, parent]:
                logger.info('Killing process {0}:\n{1}'.format(pid, line))
                os.kill(pid, signal.SIGKILL)


def main():
    fname = os.path.dirname(__file__) + "/_agent.log"
    logging.basicConfig(
        level=logging.DEBUG,
        filename=fname,
        format='%(asctime)s [%(levelname)s] %(name)s:%(lineno)d %(message)s')

    parser = ArgumentParser()
    parser.add_argument(
        "--telegraf",
        dest="telegraf_path",
        help="telegraf_path",
        default="/tmp/telegraf")
    parser.add_argument(
        "--host",
        dest="hostname_path",
        help="telegraf_path",
        default="/usr/bin/telegraf")
    parser.add_argument(
        "-k", "--kill-old",
        action="store_true",
        dest="kill_old"
    )
    options = parser.parse_args()

    logger.info('Init')
    customs_script = os.path.dirname(__file__) + '/agent_customs.sh'
    # todo: deprecate
    if options.kill_old:
        kill_old_agents(options.telegraf_path)

    try:
        logger.info('Trying to make telegraf executable: %s', options.telegraf_path)
        # 0o755 compatible with old python versions. 744 is NOT enough
        os.chmod(options.telegraf_path, 493)
    except OSError:
        logger.warning(
            'Unable to set %s access rights to execute.',
            options.telegraf_path,
            exc_info=True)
    try:
        logger.info('Trying to make customs script executable: %s', customs_script)
        # 0o755 compatible with old python versions. 744 is NOT enough
        os.chmod(customs_script, 493)
    except OSError:
        logger.warning(
            'Unable to set %s access rights to execute.',
            customs_script,
            exc_info=True)

    worker = AgentWorker(options.telegraf_path)
    worker.read_startup_config()

    logger.info('Starting AgentWorker: %s', worker)
    worker.start()

    try:
        logger.debug("Check for any stdin command for shutdown")
        cmd = sys.stdin.readline()
        if cmd:
            logger.info("Stdin cmd received: %s", cmd)
    except KeyboardInterrupt:
        logger.debug("Interrupted")
    except BaseException:
        logger.exception("Something nasty happened while waiting for stop")
    worker.finished = True
    agent_finished = False
    while not agent_finished:
        try:
            if worker.is_alive():
                logger.debug("Join the worker thread, waiting for cleanup")
                worker.join(10)
            if worker.is_alive():
                logger.error("Worker have not finished shutdown in 10 seconds, going to exit anyway")
                worker.kill()
                agent_finished = True
            else:
                agent_finished = True
        except BaseException:
            logger.exception("Something nasty happened while waiting for worker shutdown")


if __name__ == '__main__':
    set_sig_handler()
    main()



import logging
import requests
import threading
import time
import datetime
import os
import pkg_resources
try:
    import queue
except ImportError:
    import Queue as queue

from retrying import retry, RetryError

from ..common.interfaces import AbstractClient
from ..common.util import pretty_print, thread_safe_property

from requests.packages.urllib3.exceptions import InsecureRequestWarning
requests.packages.urllib3.disable_warnings(InsecureRequestWarning)


logger = logging.getLogger(__name__)

RETRY_ARGS = dict(
    wrap_exception=True,
    stop_max_delay=10000,
    wait_fixed=1000,
    stop_max_attempt_number=5
)


@retry(**RETRY_ARGS)
def send_chunk(session, req, timeout=5):
    r = session.send(req, verify=False, timeout=timeout)
    logger.debug('Request %s code %s. Text: %s', r.url, r.status_code, r.text)
    r.raise_for_status()
    return r


class LunaparkVoltaClient(AbstractClient):
    create_job_path = '/mobile/create_job.json'
    update_job_path = '/mobile/update_job.json'
    metric_update = ''  # FIXME
    metric_upload = '/api/volta/?query='
    dbname = 'volta'
    symlink_artifacts_path = 'lunapark_volta'
    data_types_to_tables = {
        'current': 'currents',
        'syncs': 'syncs',
        'events': 'events',
        'metrics': 'metrics',
        'fragments': 'fragments',
        'unknown': 'logentries'
    }
    # aliases
    data_types_to_tables.update(
        {
            'currents': data_types_to_tables['current'],
            'sync': data_types_to_tables['syncs'],
            'event': data_types_to_tables['events'],
            'metric': data_types_to_tables['metrics'],
            'fragment': data_types_to_tables['fragments'],
            'logentries': data_types_to_tables['unknown']
        }
    )
    clickhouse_output_fmt = {
        'current': ['ts', 'value'],
        'syncs': ['ts', 'log_uts', 'app', 'tag', 'message'],
        'events': ['ts', 'log_uts', 'app', 'tag', 'value'],
        'metrics': ['ts', 'log_uts', 'app', 'tag', 'value'],
        'fragments': ['ts', 'log_uts', 'app', 'tag', 'message'],
        'unknown': ['ts', 'value']
    }
    # aliases
    clickhouse_output_fmt.update(
        {
            'currents': clickhouse_output_fmt['current'],
            'sync': clickhouse_output_fmt['syncs'],
            'event': clickhouse_output_fmt['events'],
            'metric': clickhouse_output_fmt['metrics'],
            'fragment': clickhouse_output_fmt['fragments'],
            'logentries': clickhouse_output_fmt['unknown']
        }
    )

    def __init__(self, meta, job):
        super(LunaparkVoltaClient, self).__init__(meta, job)
        self.failed = threading.Event()
        if self.meta.get('api_address'):
            self.api_address = self.meta.get('api_address')
        else:
            raise RuntimeError('Api address must be specified')
        self.clickhouse_cols = ['key_date', 'test_id']
        self.task = self.meta.get('task', 'LOAD-272')
        self.session = requests.session()
        self.key_date = "{key_date}".format(key_date=datetime.datetime.now().strftime("%Y-%m-%d"))
        self._job_number = None
        self.worker = WorkerThread(self)
        self.worker.start()
        logger.info('Lunapark Volta public job id: %s', self.job_number)

    @thread_safe_property
    def job_number(self):
        if self.failed.is_set():
            return
        if not self._job_number:
            try:
                self._job_number = self.create_job()
                self.__test_id_link_to_jobno()
            except RetryError:
                logger.debug('Failed to create lunapark volta job', exc_info=True)
                logger.warning('Failed to create lunapark volta job')
                self.failed.set()
            else:
                return self._job_number
        else:
            return self._job_number

    def __test_id_link_to_jobno(self):
        """  create symlink local_id <-> public_id  """
        link_dir = os.path.join(self.data_session.artifacts_base_dir, self.symlink_artifacts_path)
        if not self._job_number:
            logger.info('Public test id not available, skipped symlink creation for %s', self.symlink_artifacts_path)
            return
        if not os.path.exists(link_dir):
            os.makedirs(link_dir)
        try:
            os.symlink(
                os.path.join(
                    os.path.relpath(self.data_session.artifacts_base_dir, link_dir), self.data_session.job_id
                ),
                os.path.join(link_dir, str(self.job_number))
            )
        except OSError:
            logger.warning(
                'Unable to create %s/%s symlink for test: %s',
                self.symlink_artifacts_path, self.job_number, self.data_session.job_id
            )
        else:
            logger.debug(
                'Symlink %s/%s created for job: %s', self.symlink_artifacts_path, self.job_number, self.data_session.job_id
            )

    def put(self, data_type, df):
        if not self.failed.is_set():
            self.pending_queue.put((data_type, df))
        else:
            logger.debug('Skipped incoming data chunk due to failures')

    def create_job(self):
        my_user_agent = None
        try:
            my_user_agent = pkg_resources.require('yandextank.contrib.netort.netort')[0].version
        except pkg_resources.DistributionNotFound:
            my_user_agent = 'DistributionNotFound'
        finally:
            headers = {
                "User-Agent": "Uploader/{uploader_ua}, {upward_ua}".format(
                    upward_ua=self.meta.get('user_agent', ''),
                    uploader_ua=my_user_agent
                )
            }
        req = requests.Request(
            'POST',
            "{api_address}{path}".format(
                api_address=self.api_address,
                path=self.create_job_path
            ),
            headers=headers
        )
        req.data = {
            'key_date': self.key_date,
            'test_id': "{key_date}_{local_job_id}".format(
                key_date=self.key_date,
                local_job_id=self.data_session.job_id
            ),
            'task': self.task,
            'version': "2"
        }
        prepared_req = req.prepare()
        logger.debug('Prepared lunapark_volta create_job request:\n%s', pretty_print(prepared_req))

        response = send_chunk(self.session, prepared_req)
        logger.debug('Lunapark volta create job status: %s', response.status_code)
        logger.debug('Answ data: %s', response.json())
        job_id = response.json().get('jobno')
        if not job_id:
            logger.warning('Create job answ data: %s', response.json())
            self.failed.set()
            raise ValueError('Lunapark volta returned answer without jobno: %s', response.json())
        else:
            logger.info('Lunapark volta job created: %s', job_id)
            return job_id

    def update_job(self, meta):
        req = requests.Request(
            'POST',
            "{api_address}{path}".format(
                api_address=self.api_address,
                path=self.update_job_path
            ),
        )
        req.data = meta
        req.data['test_start'] = self.data_session.test_start
        req.data['test_id'] = "{key_date}_{local_job_id}".format(
            key_date=self.key_date,
            local_job_id=self.data_session.job_id
        ),
        prepared_req = req.prepare()
        logger.debug('Prepared update_job request:\n%s', pretty_print(prepared_req))
        response = send_chunk(self.session, prepared_req)
        logger.debug('Update job status: %s', response.status_code)
        logger.debug('Answ data: %s', response.content)

    def get_info(self):
        """ mock """
        pass

    def close(self, test_end):
        self.worker.stop()
        while not self.worker.is_finished():
            logger.debug('Processing pending uploader queue... qsize: %s', self.pending_queue.qsize())
        logger.debug('Joining lunapark_volta metric uploader thread...')
        self.worker.join()


class WorkerThread(threading.Thread):
    """ Process data """
    def __init__(self, client):
        super(WorkerThread, self).__init__()
        self._finished = threading.Event()
        self._interrupted = threading.Event()
        self.client = client
        self.session = requests.session()

    def run(self):
        while not self._interrupted.is_set():
            self.__process_pending_queue()
        logger.info('Lunapark volta uploader thread main loop interrupted, '
                    'finishing work and trying to send the rest of data, qsize: %s',
                    self.client.pending_queue.qsize())
        self.__process_pending_queue()
        self._finished.set()

    def __process_pending_queue(self):
        exec_time_start = time.time()
        try:
            incoming_df = self.client.pending_queue.get_nowait()
            df = incoming_df.copy()
        except queue.Empty:
            time.sleep(1)
        else:
            for metric_local_id, df_grouped_by_id in df.groupby(level=0, sort=False):
                df_grouped_by_id['key_date'] = self.client.key_date
                df_grouped_by_id['test_id'] = "{key_date}_{local_job_id}".format(
                    key_date=self.client.key_date,
                    local_job_id=self.client.data_session.job_id
                )
                metric = self.client.data_session.manager.get_metric_by_id(metric_local_id)
                if metric.type == 'events':
                    try:
                        gb = df_grouped_by_id.groupby('custom_metric_type', sort=False)
                    except KeyError:
                        self.__send_this_type(df_grouped_by_id, 'unknown')
                    else:
                        for gb_name, gb_frame in gb:
                            self.__send_this_type(gb_frame, gb_name)
                elif metric.type == 'metrics' and metric.meta.get('name') == 'current':
                    self.__send_this_type(df_grouped_by_id, 'current')
                else:
                    logger.debug('Dropped data of type %s, %s probably monitoring or other user custom metrics...',
                                 metric.type, metric.meta.get('name'))
        logger.debug('Lunapark volta client processing took %.2f ms', (time.time() - exec_time_start) * 1000)

    def __send_this_type(self, df, metric_type):
        try:
            body = df.to_csv(
                sep='\t',
                header=False,
                index=False,
                na_rep="",
                columns=self.client.clickhouse_cols + self.client.clickhouse_output_fmt[metric_type]
            )
        except Exception:
            logger.info('Exc: %s', exc_info=True)
        req = requests.Request(
            'POST', "{api}{data_upload_handler}{query}".format(
                api=self.client.api_address,
                data_upload_handler=self.client.metric_upload,
                query="INSERT INTO {table} FORMAT TSV".format(
                    table="{db}.{type}".format(
                        db=self.client.dbname,
                        type=self.client.data_types_to_tables[metric_type]
                    )
                )
            )
        )
        req.data = body
        prepared_req = req.prepare()
        try:
            send_chunk(self.session, prepared_req)
        except RetryError:
            logger.warning(
                'Failed to upload data to lunapark backend. Dropped some data.'
            )
            logger.debug(
                'Failed to upload data to luna, dropped data: %s', df, exc_info=True
            )
            return

    def is_finished(self):
        return self._finished

    def stop(self):
        self._interrupted.set()
        # FIXME
        logger.info('Lunapark Volta public job id: http://lunapark.yandex-team.ru/mobile/%s', self.client.job_number)



import queue
import hashlib
import logging
import os
import subprocess
import tempfile
import threading
import time
from shutil import copyfile, rmtree
from ...common.util import SecuredShell

from ..Telegraf.config import AgentConfig, create_agent_py
from ..Telegraf.reader import MonitoringReader

logger = logging.getLogger(__name__)


def generate_file_md5(filename, blocksize=2 ** 20):
    m = hashlib.md5()
    with open(filename, "rb") as f:
        while True:
            buf = f.read(blocksize)
            if not buf:
                break
            m.update(buf)
    return m.hexdigest()


class LocalhostClient(object):
    """ localhost client setup """
    AGENT_FILENAME = 'agent.py'

    def __init__(self, config, old_style_configs, kill_old):
        # config
        self.kill_old = '--kill-old' if kill_old else ''
        self.python = config['python']
        self.host = "localhost"
        self.telegraf = config['telegraf']
        self.config = AgentConfig(config, old_style_configs)

        # connection
        self.incoming_queue = queue.Queue()
        self.buffer = ""

        self.workdir = None
        self.reader = MonitoringReader(self.incoming_queue)

        self.path = {
            'AGENT_LOCAL_PATH': create_agent_py(self.AGENT_FILENAME),
            'TELEGRAF_LOCAL_PATH': self.telegraf,
        }

    def install(self):
        self.workdir = tempfile.mkdtemp()
        logger.info("Created temp dir %s", self.workdir)
        agent_config = self.config.create_collector_config(self.workdir)
        startup_config = self.config.create_startup_config()
        customs_script = self.config.create_custom_exec_script()
        try:
            copyfile(
                self.path['AGENT_LOCAL_PATH'],
                os.path.join(
                    self.workdir,
                    self.AGENT_FILENAME))
            copyfile(agent_config, os.path.join(self.workdir, 'agent.cfg'))
            copyfile(startup_config, os.path.join(
                self.workdir,
                'agent_startup.cfg'))
            copyfile(
                customs_script,
                os.path.join(
                    self.workdir,
                    'agent_customs.sh'))
            if not os.path.isfile(self.path['TELEGRAF_LOCAL_PATH']):
                logger.error(
                    'Telegraf binary not found at specified path: %s\n'
                    'You can find telegraf binaries here: https://github.com/influxdata/telegraf\n'
                    'or install debian package: `telegraf`',
                    self.path['TELEGRAF_LOCAL_PATH'])
                return None, None, None
        except Exception:
            logger.error("Failed to copy agent to %s on localhost", self.workdir)
            logger.debug("Failed to copy agent to %s on localhost", self.workdir, exc_info=True)
            return None, None, None
        return agent_config, startup_config, customs_script

    @staticmethod
    def popen(args):
        return subprocess.Popen(
            args,
            bufsize=0,
            preexec_fn=os.setsid,
            close_fds=True,
            shell=False,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            stdin=subprocess.PIPE, )

    def start(self):
        """Start local agent"""
        args = self.python.split() + [
            os.path.join(
                self.workdir,
                self.AGENT_FILENAME),
            '--telegraf',
            self.path['TELEGRAF_LOCAL_PATH'],
            '--host',
            self.host]
        if self.kill_old:
            args.append(self.kill_old)
        logger.info('Starting agent on localhost: {}'.format(args))
        self.session = self.popen(args)
        self.reader_thread = threading.Thread(target=self.read_buffer)
        self.reader_thread.daemon = True
        return self.session

    def read_buffer(self):
        while self.session:
            try:
                chunk = self.session.stdout.read(4096).decode('utf8')
                if chunk:
                    parts = chunk.rsplit('\n', 1)
                    if len(parts) > 1:
                        ready_chunk = self.buffer + parts[0] + '\n'
                        self.buffer = parts[1]
                        self.incoming_queue.put(ready_chunk)
                    else:
                        self.buffer += parts[0]
            except ValueError:
                logger.debug(
                    'this exc most likely raised during interpreter shutdown\n'
                    'otherwise something really nasty happend',
                    exc_info=True)

    def _stop_agent(self):
        if self.session:
            self.session.terminate()
            logger.info('Waiting localhost agent to terminate...')

    def uninstall(self):
        """
        Remove agent's files from remote host
        """
        if self.session:
            self.session.wait()
            self.session = None
        log_filename = "agent_{host}.log".format(host="localhost")
        data_filename = "agent_{host}.rawdata".format(host="localhost")
        try:
            logger.info('Saving monitoring artifacts from localhost')
            copyfile(self.workdir + "/_agent.log", log_filename)
            copyfile(self.workdir + "/monitoring.rawdata", data_filename)
            logger.info('Deleting temp directory: %s', self.workdir)
            rmtree(self.workdir)
        except Exception:
            logger.error("Exception while uninstalling agent", exc_info=True)

        logger.info("Removing agent from: localhost")
        return log_filename, data_filename


class SSHClient(object):
    """remote agent client setup """
    AGENT_FILENAME = 'agent.py'

    def __init__(self, config, old_style_configs, timeout, kill_old):
        # config
        self.kill_old = '--kill-old' if kill_old else ''
        self.host = config['host']
        self.username = config['username']
        self.python = config['python']
        self.ssh_key_path = config['ssh_key_path']
        self.port = config['port']
        self.telegraf = config['telegraf']
        self.config = AgentConfig(config, old_style_configs)

        # connection
        self.session = None
        self.ssh = SecuredShell(
            self.host, self.port, self.username, command_timeout=timeout, ssh_key_path=self.ssh_key_path
        )
        self.incoming_queue = queue.Queue()
        self.buffer = ""
        self.stop_sent = None
        self.successfull_stop = None

        self.reader = MonitoringReader(self.incoming_queue)
        handle, cfg_path = tempfile.mkstemp('.cfg', 'agent_')
        os.close(handle)
        self.path = {
            # Destination path on remote host
            'AGENT_REMOTE_FOLDER': '/tmp/',
            # Source path on tank
            'AGENT_LOCAL_PATH': create_agent_py(self.AGENT_FILENAME),
            'TELEGRAF_REMOTE_PATH': '/tmp/telegraf',
            'TELEGRAF_LOCAL_PATH': self.telegraf,
        }
        self.agent_remote_folder = None

    def install(self):
        """Create folder and copy agent and metrics scripts to remote host"""
        logger.info('Trying to establish SSH connection with %s', self.host)
        try:
            self.ssh.ensure_connection()
        except Exception:
            logger.error('Failed to establish SSH connection with %s', self.host)
            return None, None, None

        logger.info("Installing monitoring agent at %s@%s...", self.username, self.host)

        # create remote temp dir
        cmd = 'mktemp -d'
        logger.info("Creating temp dir on %s", self.host)
        try:
            out, errors, err_code = self.ssh.execute(cmd)
        except Exception:
            logger.exception("Failed to install monitoring agent to %s", self.host)
            return None, None, None

        if err_code:
            logger.error(
                "Failed to create remote dir via SSH at %s@%s, code %s: %s" %
                (self.username, self.host, err_code, out.strip()))
            logger.error("[%s] error: '%s'", self.host, errors)
            logger.error("Cancelling agent installation on %s", self.host)
            return None, None, None

        remote_dir = out.strip()
        if remote_dir:
            self.path['AGENT_REMOTE_FOLDER'] = remote_dir
            self.agent_remote_folder = remote_dir
        logger.info("Remote dir at %s:%s", self.host, self.path['AGENT_REMOTE_FOLDER'])

        # create collector config
        agent_config = self.config.create_collector_config(self.path['AGENT_REMOTE_FOLDER'])
        startup_config = self.config.create_startup_config()
        customs_script = self.config.create_custom_exec_script()

        # trying to detect get information about telegraf client
        cmd = 'stat {path}'.format(path=self.path['TELEGRAF_REMOTE_PATH'])
        remote_telegraf_exists = False
        try:
            out, err, err_code = self.ssh.execute(cmd)
        except Exception:
            logger.exception("SSH execute error trying to check telegraf availability on host %s", self.host)
        else:
            if not err_code:
                remote_telegraf_exists = True

        try:
            if remote_telegraf_exists:
                logger.info('Found telegraf client on %s..', self.host)
            else:
                logger.info('Not found telegraf client on %s, trying to install from tank. Copying..', self.host)
                if os.path.isfile(self.path['TELEGRAF_LOCAL_PATH']):
                    self.ssh.send_file(self.path['TELEGRAF_LOCAL_PATH'], self.path['TELEGRAF_REMOTE_PATH'])
                elif os.path.isfile("/usr/bin/telegraf"):
                    self.ssh.send_file('/usr/bin/telegraf', self.path['TELEGRAF_REMOTE_PATH'])
                else:
                    logger.error(
                        'Telegraf binary not found neither on %s nor on localhost at specified path: %s\n'
                        'You can find telegraf binaries here: https://github.com/influxdata/telegraf\n'
                        'or install debian package: `telegraf`', self.host, self.path['TELEGRAF_LOCAL_PATH'])
                    return None, None, None

            self.ssh.send_file(
                self.path['AGENT_LOCAL_PATH'],
                os.path.join(self.path['AGENT_REMOTE_FOLDER'], self.AGENT_FILENAME))
            self.ssh.send_file(
                agent_config,
                os.path.join(self.path['AGENT_REMOTE_FOLDER'], 'agent.cfg'))
            self.ssh.send_file(
                startup_config,
                os.path.join(self.path['AGENT_REMOTE_FOLDER'], 'agent_startup.cfg'))
            self.ssh.send_file(
                customs_script,
                os.path.join(self.path['AGENT_REMOTE_FOLDER'], 'agent_customs.sh'))

        except Exception:
            logger.exception("Failed to install agent on %s", self.host)
            return None, None, None

        return agent_config, startup_config, customs_script

    def start(self):
        """Start remote agent"""
        logger.info('Starting agent: %s', self.host)
        command = "{python} {agent_path} --telegraf {telegraf_path} --host {host} {kill_old}".format(
            python=self.python,
            agent_path=os.path.join(
                self.path['AGENT_REMOTE_FOLDER'],
                self.AGENT_FILENAME),
            telegraf_path=self.path['TELEGRAF_REMOTE_PATH'],
            host=self.host,
            kill_old=self.kill_old)
        logger.debug('Command to start agent: %s', command)
        self.session = self.ssh.async_session(command)
        self.reader_thread = threading.Thread(target=self.read_buffer)
        self.reader_thread.daemon = True
        return self.session

    def read_buffer(self):
        while self.session and not self.stop_sent:
            chunk = self.session.read_maybe()
            if chunk:
                parts = chunk.rsplit('\n', 1)
                if len(parts) > 1:
                    ready_chunk = self.buffer + parts[0] + '\n'
                    self.buffer = parts[1]
                    self.incoming_queue.put(ready_chunk)
                else:
                    self.buffer += parts[0]
            else:
                time.sleep(1)
        logger.info('Daemon reader stopped')

    def _stop_agent(self):
        """
        Stop data collection. Separated from uninstall to speed up multihost processing
        """

        try:
            if self.session:
                self.session.send(b"stop\n")
                self.stop_sent = time.time()
        except BaseException:
            logger.warning(
                'Unable to correctly stop monitoring agent - session is broken on %s.',
                self.host,
                exc_info=True)

    def uninstall(self):
        """
        Remove agent's files from remote host
        """
        log_filename = "agent_{host}.log".format(host=self.host)
        data_filename = "agent_{host}.rawdata".format(host=self.host)

        try:
            if self.session:
                self._wait_for_stop()
                self.session.close()
                self.session = None
        except BaseException:
            logger.warning(
                'Unable to correctly stop monitoring agent - session is broken. Pay attention to agent log (%s).',
                log_filename,
                exc_info=True)
        try:
            self.ssh.get_file(
                os.path.join(
                    self.path['AGENT_REMOTE_FOLDER'],
                    "_agent.log"),
                log_filename)
            self.ssh.get_file(
                os.path.join(
                    self.path['AGENT_REMOTE_FOLDER'],
                    "monitoring.rawdata"),
                data_filename)
            self.ssh.rm_r(self.path['AGENT_REMOTE_FOLDER'])
        except Exception:
            logger.error("Unable to get agent artefacts", exc_info=True)
        if not self.successfull_stop:
            self._kill_agent()
        return log_filename, data_filename

    def _wait_for_stop(self):
        done = False
        agent_stop_timeout = 5
        while not done:
            if not self.stop_sent:
                logger.info('Session was broken on %s, switch to kill', self.host)
                break
            if (self.stop_sent + agent_stop_timeout) < time.time():
                logger.info("Agent hasn't finished in %s sec, force quit on %s",
                            agent_stop_timeout,
                            self.host)
                break
            if self.session.is_finished():
                logger.debug('Session ended with status %s with %s',
                             self.session.exit_status(),
                             self.host)
                self.successfull_stop = True
                done = True

    def _kill_agent(self):
        if self.agent_remote_folder:
            tpl = ('main_p=$(pgrep -f "[p]ython.*{folder}");'
                   'tlgrf_p=$(pgrep -f "[t]elegraf.*{folder}");'
                   'descendent_pids(){{ if [ "x$1" != "x" ]; then '
                   'pids=$(pgrep -P $1); echo $pids;'
                   'for p in $pids; do descendent_pids $p; done; fi }};'
                   'all_p=$(descendent_pids ${{main_p}});'
                   'if [ "x${{main_p}}${{tlgrf_p}}${{all_p}}" != "x" ] ; then '
                   ' kill -9 ${{main_p}} ${{tlgrf_p}} ${{all_p}}; fi')
            cmd = tpl.format(folder=self.agent_remote_folder)
            out, errors, err_code = self.ssh.execute(cmd)
            if errors:
                logger.error(
                    "[%s] error while killing agent: '%s'",
                    self.host,
                    errors)



import logging
import time
import typing
from yandextank.common.interfaces import MonitoringCollectorProtocol, MonitoringChunk
from yandextank.common.util import observetime
from re import match, findall
from threading import Thread, Event, Lock
from queue import Queue

Metric = typing.Dict[str, typing.Any]
Buffer = typing.Dict[float, typing.Dict[str, typing.Any]]
Composer = typing.Callable[[Metric, Buffer], None]

LOGGER = logging.getLogger(__name__)


def convert_name(name: str):
    NAME_MAP = {r'conv\(.+\)': r'(?<=conv\().+?(?=,)'}
    for pattern, mask in NAME_MAP.items():
        if match(pattern, name):
            name = findall(mask, name)[0]
            break
    return 'custom:{}'.format(name)[:100]


def monitoring_data(pname, metrics, comment=''):
    try:
        return {
            'timestamp': tuple(metrics)[0],
            'data': {
                pname: {
                    'comment': comment,
                    'metrics': {convert_name(name): value for name, value in tuple(metrics.values())[0].items()},
                }
            },
        }
    except (IndexError, ValueError) as ve:
        LOGGER.warning("Can't format. Wrong data %s. %s", metrics, ve, exc_info=True)


class MonitoringPanel(object):

    '''
    The class Panel is used for initializing and managing structural levels for collecting monitoring data passed through the configuration file.
    '''

    def __init__(self, name, timeout, queue: Queue):
        self.name = name
        self.queue = queue
        self.timeout = timeout
        self.sensors = set()
        self.buffer: Buffer = dict()

    def add_sensors(self, sensors):
        self.sensors.union(sensors)

    @observetime('MonitoringPanel.process_queue', LOGGER)
    def process_queue(self):
        try:
            while not self.queue.empty():
                data = self.queue.get()
                metrics = []
                if isinstance(data, dict) and 'timestamp' in data:
                    metrics = [data]
                elif isinstance(data, list):
                    metrics = list(filter(lambda d: isinstance(d, typing.Mapping), data))
                for m in metrics:
                    ts = m['timestamp']
                    if ts in self.buffer:
                        self.buffer[ts].update({m['sensor']: m['value']})
                    else:
                        self.buffer[ts] = {m['sensor']: m['value']}
        except KeyError as error:
            LOGGER.warning('Got data in invalid format %s', error)
        except (IOError, OSError) as error:
            LOGGER.warning('Panel %s processing queue error. %s', self.name, error, exc_info=True)

    def get_metrics(self):
        now = time.time()
        outbuffer = list()
        complete_ts = list()
        for ts, metrics in self.buffer.items():
            if self.sensors.issubset(set(metrics.keys())) or (now - ts) > self.timeout:
                outbuffer.append({ts: metrics})
                complete_ts.append(ts)
        for ts in complete_ts:
            self.buffer.pop(ts)
        return outbuffer


class MonitoringSensorProtocol(typing.Protocol):
    def fetch_metrics(self):
        ...


class DefaultCollector(MonitoringCollectorProtocol):
    def __init__(self, *, logger: logging.Logger, timeout: float, poll_interval: float):
        self.logger = logger
        self.stop_event = Event()
        self.delivery_interval = timeout
        self.poll_interval = poll_interval
        self.sensors: typing.List[Thread] = []
        self.panels: typing.List[Thread] = []
        self._data_buffer = []
        self._buffer_lock = Lock()

    def add_panel(self, panel: MonitoringPanel):
        self.panels.append(Thread(target=self.run_panel, args=(panel,)))

    def add_sensor(self, sensor: MonitoringSensorProtocol):
        self.sensors.append(Thread(target=self.run_sensor, args=(sensor,)))

    def start(self):
        self.logger.debug('Starting %s panels', len(self.panels))
        for thread in self.panels:
            if not thread.is_alive():
                thread.start()
        self.logger.debug('Starting %s sensors', len(self.sensors))
        for thread in self.sensors:
            if not thread.is_alive():
                thread.start()

    def stop(self):
        self.logger.info('Stopping Monitoring Collector in %s seconds', self.poll_interval)
        time.sleep(self.poll_interval + 1)
        self.stop_event.set()
        for thread in self.sensors:
            if thread.is_alive():
                thread.join()
        for thread in self.panels:
            if thread.is_alive():
                thread.join()

    def poll(self) -> MonitoringChunk:
        with self._buffer_lock:
            data, self._data_buffer = self._data_buffer, []
        return data

    def run_sensor(self, sensor: MonitoringSensorProtocol):
        while not self.stop_event.wait(self.poll_interval):
            sensor.fetch_metrics()

    def run_panel(self, panel: MonitoringPanel):
        try:
            while not self.stop_event.wait(self.poll_interval):
                panel.process_queue()
                with self._buffer_lock:
                    for metric in panel.get_metrics():
                        self._data_buffer.append(monitoring_data(panel.name, metric))
        except Exception as ex:
            self.logger.exception('Error %s for monitoring panel %s.', type(ex), panel.name)
        finally:
            self.logger.info('Monitoring panel %s will be closed after %s second', panel.name, self.delivery_interval)
            time.sleep(self.delivery_interval)
            panel.process_queue()
            with self._buffer_lock:
                for ts, metrics in panel.buffer.items():
                    self._data_buffer.append(monitoring_data(panel.name, {ts: metrics}))



from collections import OrderedDict

from requests import HTTPError, ConnectionError
from requests.exceptions import Timeout, TooManyRedirects

from ..common.interfaces import AbstractClient, QueueWorker
from ..common.util import pretty_print, thread_safe_property

from retrying import retry

import json
import pkg_resources
import logging
import requests
import threading
import time
import six
import pandas as pd
import datetime
import os
if six.PY2:
    import Queue as queue
else:
    # noinspection PyUnresolvedReferences
    import queue

requests.packages.urllib3.disable_warnings()

logger = logging.getLogger(__name__)


RETRY_ARGS = dict(
    stop_max_delay=30000,
    wait_fixed=3000,
    stop_max_attempt_number=10
)

SLEEP_ON_EMPTY = 0.2  # pause in seconds before checking empty queue on new items
MAX_DF_LENGTH = 10000  # Size of chunk is 10k rows, it's approximately 0.5Mb in csv


@retry(**RETRY_ARGS)
def send_chunk(session, req, timeout=5):
    r = session.send(req, verify=False, timeout=timeout)
    logger.debug('Request %s code %s. Text: %s', r.url, r.status_code, r.text)
    return r


def if_not_failed(func):
    def wrapped(self, *a, **kw):
        if self.failed.is_set():
            logger.warning('Luna client is disabled')
            return
        else:
            return func(self, *a, **kw)
    return wrapped


class LunaClient(AbstractClient):
    create_metric_path = '/create_metric/'
    update_metric_path = '/update_metric/'
    upload_metric_path = '/upload_metric/?query='  # production
    create_job_path = '/create_job/'
    update_job_path = '/update_job/'
    close_job_path = '/close_job/'
    symlink_artifacts_path = 'luna'

    def __init__(self, meta, job):
        super(LunaClient, self).__init__(meta, job)
        logger.debug('Luna client local id: %s', self.local_id)
        self.dbname = meta.get('db_name', 'luna')
        self.failed = threading.Event()
        self.public_ids = {}
        self.luna_columns = ['key_date', 'tag']
        self.key_date = "{key_date}".format(key_date=datetime.datetime.now().strftime("%Y-%m-%d"))
        self._interrupted = threading.Event()
        self.register_worker = RegisterWorkerThread(self)
        self.register_worker.start()
        self.worker = WorkerThread(self)
        self.worker.start()
        self.session = requests.session()
        self.clickhouse_user = meta.get('clickhouse_user', 'lunapark')
        self.clickhouse_key = meta.get('clickhouse_key', 'lunapark')
        self.max_df_len = meta.get('max_df_len') or MAX_DF_LENGTH

        if self.meta.get('api_address'):
            self.api_address = self.meta.get('api_address')
        else:
            raise RuntimeError('Api address SHOULD be specified')
        self._job_number = None

    @thread_safe_property
    def job_number(self):
        if not self.failed.is_set():
            try:
                _job_number = self.create_job()
                self.__test_id_link_to_jobno(_job_number)
            except (HTTPError, ConnectionError, Timeout, TooManyRedirects):
                logger.error('Failed to create Luna job', exc_info=True)
                self.failed.set()
                self.worker.interrupt()
                self.interrupt()
            else:
                return _job_number

    def put(self, data_type, df):   # noqa: PLE0202
        if not self.failed.is_set():
            self.pending_queue.put((data_type, df))
        else:
            logger.debug('Skipped incoming data chunk due to failures')

    @if_not_failed
    def create_job(self):
        """ Create public Luna job

        Returns:
            job_id (str): Luna job id
        """
        my_user_agent = None
        try:
            my_user_agent = pkg_resources.require('yandextank.contrib.netort.netort')[0].version
        except pkg_resources.DistributionNotFound:
            my_user_agent = 'DistributionNotFound'
        finally:
            headers = {
                "User-Agent": "Uploader/{uploader_ua}, {upward_ua}".format(
                    upward_ua=self.meta.get('user_agent', ''),
                    uploader_ua=my_user_agent
                )
            }
        request_data = dict(
            self.data_session.tankapi_info,
            test_start=int(self.data_session.test_start),
            configs=self._get_encoded_configs_content()
        )
        if request_data['host'] is None or request_data['port'] is None:
            logger.warning('Tankapi host and/or port are unspecified. Artifacts & configs would be unavailable in Luna')
        req = requests.Request(
            'POST',
            "{api_address}{path}".format(
                api_address=self.api_address,
                path=self.create_job_path
            ),
            headers=headers,
            data=request_data
        )
        prepared_req = req.prepare()
        logger.debug('Prepared create_job request:\n%s', pretty_print(prepared_req))

        response = send_chunk(self.session, prepared_req)
        logger.debug('Luna create job status: %s', response.status_code)
        response.raise_for_status()
        logger.debug('Answ data: %s', response.content)
        job_id = response.content.decode('utf-8') if isinstance(response.content, bytes) else response.content
        if not job_id:
            self.failed.set()
            self.worker.interrupt()
            raise ValueError('Luna returned answer without jobid: %s', response.content)
        else:
            logger.info('Luna job created: %s', job_id)
            return job_id

    def _get_encoded_configs_content(self):
        config_dir = self.data_session.artifacts_base_dir
        config_filenames = self.data_session.config_filenames
        files = config_filenames & set(os.listdir(config_dir))
        result = {}
        for file_name in files:
            with open(os.path.join(config_dir, file_name)) as file:
                content = file.read()
                result[file_name] = content
        return json.dumps(result)

    @if_not_failed
    def update_job(self, meta):
        req = requests.Request(
            'POST',
            "{api_address}{path}?job={job}".format(
                api_address=self.api_address,
                path=self.update_job_path,
                job=self.job_number
            ),
            json=meta
        )
        prepared_req = req.prepare()
        logger.debug('Prepared update_job request:\n%s', pretty_print(prepared_req))
        response = send_chunk(self.session, prepared_req)
        logger.debug('Update job status: %s', response.status_code)
        logger.debug('Answ data: %s', response.content)

    @if_not_failed
    def update_metric(self, meta):
        for metric_tag, metric_obj in self.data_session.manager.metrics.items():
            if not metric_obj.tag:
                logger.debug('Metric %s has no public tag, skipped updating metric', metric_tag)
                continue
            req = requests.Request(
                'POST',
                "{api_address}{path}?tag={tag}".format(
                    api_address=self.api_address,
                    path=self.update_metric_path,
                    tag=metric_obj.tag
                ),
            )
            req.data = meta
            # FIXME: should be called '_offset' after volta-service production is updated;
            if 'sys_uts_offset' in meta and metric_obj.type == 'metrics':
                req.data['offset'] = meta['sys_uts_offset']
            elif 'log_uts_offset' in meta and metric_obj.type == 'events':
                req.data['offset'] = meta['log_uts_offset']
            prepared_req = req.prepare()
            logger.debug('Prepared update_metric request:\n%s', pretty_print(prepared_req))
            response = send_chunk(self.session, prepared_req)
            logger.debug('Update metric status: %s', response.status_code)
            logger.debug('Answ data: %s', response.content)

    @if_not_failed
    def _close_job(self, duration):
        req = requests.Request(
            'GET',
            "{api_address}{path}".format(
                api_address=self.api_address,
                path=self.close_job_path,
            ),
            params={'job': self.job_number,
                    'duration': int(duration)}
        )
        prepared_req = req.prepare()
        logger.debug('Prepared close_job request:\n%s', pretty_print(prepared_req))
        response = send_chunk(self.session, prepared_req)
        logger.debug('Update job status: %s', response.status_code)

    def __test_id_link_to_jobno(self, jobno):
        """  create symlink local_id <-> public_id  """
        # TODO: fix symlink to local_id <-> luna_id
        link_dir = os.path.join(self.data_session.artifacts_base_dir, self.symlink_artifacts_path)
        if not jobno:
            logger.info('Public test id not available, skipped symlink creation for %s', self.symlink_artifacts_path)
            return
        if not os.path.exists(link_dir):
            os.makedirs(link_dir)
        try:
            os.symlink(
                os.path.join(
                    os.path.relpath(self.data_session.artifacts_base_dir, link_dir), self.data_session.job_id
                ),
                os.path.join(link_dir, str(jobno))
            )
        except OSError:
            logger.warning(
                'Unable to create %s/%s symlink for test: %s',
                self.symlink_artifacts_path, jobno, self.data_session.job_id
            )
        else:
            logger.debug(
                'Symlink %s/%s created for job: %s',
                self.symlink_artifacts_path, jobno, self.data_session.job_id
            )

    def close(self, test_end):
        self.worker.stop()
        if not self.job_number:
            logger.info('Try to interrupt queue')
            self.worker.interrupt()
        if not self.worker.is_finished():
            logger.debug('Processing pending uploader queue... qsize: %s', self.pending_queue.qsize())
        logger.info('Joining luna client metric uploader thread...')
        self.worker.join()
        logger.info('Joining luna client metric registration thread...')
        self.register_worker.stop()
        self.register_worker.join()
        self._close_job(duration=test_end-self.data_session.test_start)
        # FIXME hardcoded host
        # FIXME we dont know front hostname, because api address now is clickhouse address
        logger.info('Luna job url: %s%s', 'https://luna.yandex-team.ru/tests/', self.job_number)
        logger.info('Luna client done its work')

    def interrupt(self):
        logger.warning('Luna client work was interrupted.')
        self.put = lambda *a, **kw: None
        self.register_worker.interrupt()
        self.worker.interrupt()


class RegisterWorkerThread(threading.Thread):
    """ Register metrics metadata, get public_id from luna and create map local_id <-> public_id """
    def __init__(self, client):
        """
        :type client: LunaClient
        """
        super(RegisterWorkerThread, self).__init__()
        self.client = client
        self.session = requests.session()
        self.metrics_to_register = OrderedDict()
        self.lock = threading.Lock()
        self._finished = threading.Event()
        self._stopped = threading.Event()
        self._interrupted = threading.Event()
        # Register all metrics not registered yet
        for metric_id in self.client.data_session.manager.metrics:
            if metric_id not in self.client.public_ids:
                metric = self.client.data_session.manager.get_metric_by_id(metric_id)
                self.metrics_to_register[metric_id] = metric

    def stop(self):
        self._stopped.set()

    def interrupt(self):
        self._stopped.set()
        self._interrupted.set()

    def is_finished(self):
        return self._finished.is_set()

    def register(self, metric):
        self.lock.acquire()
        self.metrics_to_register[metric.local_id] = metric
        self.lock.release()

    def run(self):
        while not self._stopped.is_set():
            self._process_pending_queue()
        while len(self.metrics_to_register) > 0 and not self._interrupted.is_set():
            self._process_pending_queue()
        if len(self.metrics_to_register) > 0:
            self.metrics_to_register.clear()
        self._finished.set()

    def _process_pending_queue(self):
        try:
            self.lock.acquire()
            local_id, metric = self.metrics_to_register.popitem(last=False)
            self.lock.release()
            if local_id not in self.client.public_ids:
                if metric.parent is not None and metric.parent.local_id not in self.client.public_ids:
                    logger.debug('Metric {} waiting for parent metric {} to be registered'.format(metric.local_id,
                                                                                                  metric.parent.local_id))
                    self.register(metric.parent)
                    self.register(metric)
                else:
                    metric.tag = self._register_metric(metric)
                    logger.debug('Successfully received tag %s for metric.local_id: %s (%s)',
                                 metric.tag, metric.local_id, metric.meta)
                    self.client.public_ids[metric.local_id] = metric.tag
        except (HTTPError, ConnectionError, Timeout, TooManyRedirects):
            self.lock.release()
            logger.error("Luna service unavailable", exc_info=True)
            self.client.interrupt()
        except KeyError:
            self.lock.release()
            time.sleep(0.5)

    def _register_metric(self, metric):
        json = {
            'job': self.client.job_number,
            'type': metric.type.table_name,
            'types': [t.table_name for t in metric.data_types],
            'local_id': metric.local_id,
            'meta': metric.meta,
            'parent': self.client.public_ids[metric.parent.local_id] if metric.parent is not None else None,
            'case': metric.case
        }
        req = requests.Request(
            'POST',
            "{api_address}{path}".format(
                api_address=self.client.api_address,
                path=self.client.create_metric_path
            ),
            json=json
        )
        prepared_req = req.prepare()
        logger.debug('Prepared create_metric request:\n%s', pretty_print(prepared_req))
        response = send_chunk(self.session, prepared_req)
        response.raise_for_status()
        if not response.content:
            raise HTTPError('Luna did not return uniq_id for metric registration: %s', response.content)
        else:
            return response.content.decode('utf-8') if six.PY3 else response.content


# noinspection PyTypeChecker
class WorkerThread(QueueWorker):
    """ Process data """
    def __init__(self, client):
        """
        :type client: LunaClient
        """
        super(WorkerThread, self).__init__(client.pending_queue)
        self.data = {'max_length': 0}
        self.client = client
        self.session = requests.session()

    def run(self):
        while not self._stopped.is_set():
            self._process_pending_queue()
        while self.queue.qsize() > 0 and not self._interrupted.is_set():
            self._process_pending_queue(progress=True)
        while self.queue.qsize() > 0:
            self.queue.get_nowait()
        if self.data['max_length'] != 0:
            self.__upload_data()
        self._finished.set()

    def _process_pending_queue(self, progress=False):
        try:
            data_type, raw_df = self.queue.get_nowait()
            if progress:
                logger.info("{} entries in queue remaining".format(self.client.pending_queue.qsize()))
        except queue.Empty:
            time.sleep(SLEEP_ON_EMPTY)
        except KeyboardInterrupt:
            self.interrupt()
        else:
            self.__update_df(data_type, raw_df)

    def __update_max_length(self, new_length):
        return new_length if self.data['max_length'] < new_length else self.data['max_length']

    def __update_df(self, data_type, df):
        metric_local_id = df['metric_local_id'].iloc[0]
        metric = self.client.data_session.manager.get_metric_by_id(metric_local_id)

        if not metric:
            logger.warning('Received unknown metric: %s! Ignored.', metric_local_id)
            return pd.DataFrame([])

        if metric.local_id not in self.client.public_ids:
            # no public_id yet, put it back
            self.client.put(data_type, df)
            self.client.register_worker.register(metric)
            return pd.DataFrame([])

        df.loc[:, 'key_date'] = self.client.key_date
        df.loc[:, 'tag'] = self.client.public_ids[metric.local_id]

        if not df.empty:
            table_name = data_type.table_name
            if not self.data.get(table_name):
                self.data[table_name] = {
                    'dataframe': df,
                    'columns': self.client.luna_columns + data_type.columns,
                }
            else:
                self.data[table_name]['dataframe'] = pd.concat([self.data[table_name]['dataframe'], df])
            self.data['max_length'] = self.__update_max_length(len(self.data[table_name]['dataframe']))

        if self.data['max_length'] >= self.client.max_df_len:
            self.__upload_data()

    def __upload_data(self):
        for table_name, data in self.data.items():
            if table_name != 'max_length':
                logger.debug('Length of data for %s is %s', table_name, len(data['dataframe']))
                try:
                    self.__send_upload(table_name, data['dataframe'], data['columns'])

                except ConnectionError:
                    logger.warning('Failed to upload data to luna backend after consecutive retries. '
                                   'Attempt to send data in two halves')
                    try:
                        self.__send_upload(
                            table_name,
                            data['dataframe'].head(len(data['dataframe']) // 2),
                            data['columns']
                        )
                        self.__send_upload(
                            table_name,
                            data['dataframe'].tail(len(data['dataframe']) - len(data['dataframe']) // 2),
                            data['columns']
                        )
                    except ConnectionError:
                        logger.warning('Failed to upload data to luna backend after consecutive retries. Sorry.')
                        return

                self.data = dict()
                self.data['max_length'] = 0

    def __send_upload(self, table_name, df, columns):
        body = df.to_csv(
            sep='\t',
            header=False,
            index=False,
            na_rep='',
            columns=columns
        )
        req = requests.Request(
            'POST', "{api}{data_upload_handler}{query}".format(
                api=self.client.api_address,  # production proxy
                data_upload_handler=self.client.upload_metric_path,
                query="INSERT INTO {db}.{table} FORMAT TSV".format(db=self.client.dbname,
                                                                   table=table_name)  # production
            )
        )
        req.data = body
        prepared_req = req.prepare()

        try:
            resp = send_chunk(self.session, prepared_req)
            resp.raise_for_status()
            logger.info('Update table %s with %s rows -- successful', table_name, df.shape[0])

        except ConnectionError:
            raise
        except (HTTPError, Timeout, TooManyRedirects) as e:
            # noinspection PyUnboundLocalVariable
            logger.warning('Failed to upload data to luna. Dropped some data.\n{}'.
                           format(resp.content if isinstance(e, HTTPError) else 'no response'))
            logger.debug(
                'Failed to upload data to luna backend after consecutive retries.\n'
                'Dropped data head: \n%s', df.head(), exc_info=True
            )
            self.client.interrupt()



import socket
from threading import Thread, Event

import pytest
from queue import Queue
from yandextank.common.util import FileScanner, FileMultiReader
from yandextank.common.util import AddressWizard, SecuredShell

from yandextank.contrib.netort.netort.data_processing import Drain, Chopper


class TestDrain(object):
    def test_run(self):
        """
        Test drain's run function (in a same thread)
        """
        source = range(5)
        destination = Queue()
        drain = Drain(source, destination)
        drain.run()
        assert destination.qsize() == 5

    def test_interrupt(self):
        """
        Test we can interrupt the drain
        """
        source = range(1000000)
        destination = Queue()
        drain = Drain(source, destination)
        drain.start()
        drain.close()
        assert destination.qsize() < 1000000

    def test_interrupt_and_wait(self):
        """
        Test we can interrupt the drain
        """
        source = range(1000000)
        destination = Queue()
        drain = Drain(source, destination)
        drain.start()
        drain.join()
        assert destination.qsize() == 1000000


class TestChopper(object):
    def test_output(self):
        source = (range(i) for i in range(5))
        expected = [0, 0, 1, 0, 1, 2, 0, 1, 2, 3]
        assert list(Chopper(source)) == expected


class TestFileScanner(object):
    @staticmethod
    def __process_chunks(chunks, sep="\n"):
        reader = FileScanner("somefile.txt", sep=sep)
        result = []
        for chunk in chunks:
            result.extend(reader._read_lines(chunk))
        return result

    def test_empty(self):
        assert self.__process_chunks([""]) == []

    def test_simple(self):
        assert self.__process_chunks(["aaa\n", "bbb\n", "ccc\n"]) == ["aaa", "bbb", "ccc"]

    def test_split(self):
        assert self.__process_chunks(["aaa\nbbb\n", "ccc\n"]) == ["aaa", "bbb", "ccc"]

    def test_join(self):
        assert self.__process_chunks(["aaa", "bbb\n", "ccc\n"]) == ["aaabbb", "ccc"]

    def test_no_first_separator(self):
        assert self.__process_chunks(["aaa"]) == []

    def test_no_last_separator(self):
        assert self.__process_chunks(["aaa\n", "bbb\n", "ccc"]) == ["aaa", "bbb"]

    def test_use_custom_separator(self):
        assert self.__process_chunks(["aaa:bbb:ccc:"], ":") == ["aaa", "bbb", "ccc"]


class TestAddressResolver(object):
    @staticmethod
    def __resolve(chunk):
        aw = AddressWizard()
        # return format: is_v6, parsed_ip, int(port), address_str
        return aw.resolve(chunk)

    def __resolve_hostname_and_test(self, address_str, test_hostname, test_port):
        passed = False
        try:
            resolved = socket.getaddrinfo(test_hostname, test_port)
        except Exception:
            # skip this check if resolver not available
            return True

        try:
            for i in resolved:
                if i[4][1] == self.__resolve(address_str)[2] and i[4][0] == self.__resolve(address_str)[1]:
                    passed = True
        except IndexError:
            pass
        assert passed

    # ipv6
    def test_ipv6(self):
        assert self.__resolve('2a02:6b8::2:242') == (True, '2a02:6b8::2:242', 80, '2a02:6b8::2:242')

    def test_ipv6_braces_port(self):
        assert self.__resolve('[2a02:6b8::2:242]:666') == (True, '2a02:6b8::2:242', 666, '2a02:6b8::2:242')

    def test_ipv6_braces_port_spaces(self):
        assert self.__resolve('[ 2a02:6b8::2:242 ]: 666') == (True, '2a02:6b8::2:242', 666, '2a02:6b8::2:242')

    def test_ipv4(self):
        assert self.__resolve('87.250.250.242') == (False, '87.250.250.242', 80, '87.250.250.242')

    def test_ipv4_port(self):
        assert self.__resolve('87.250.250.242:666') == (False, '87.250.250.242', 666, '87.250.250.242')

    def test_ipv4_braces_port(self):
        assert self.__resolve('[87.250.250.242]:666') == (False, '87.250.250.242', 666, '87.250.250.242')

    # hostname
    def test_hostname_port(self):
        self.__resolve_hostname_and_test('ya.ru:666', 'ya.ru', '666')

    def test_hostname_braces(self):
        self.__resolve_hostname_and_test('[ya.ru]', 'ya.ru', '80')

    def test_hostname_braces_port(self):
        self.__resolve_hostname_and_test('[ya.ru]:666', 'ya.ru', '666')


class TestFileMultiReader(object):
    filename = 'yandextank/common/tests/ph.out'

    @staticmethod
    def mock_consumer(f, expected, step, errors):
        for line in [expected[i: i + step] for i in range(0, len(expected), step)]:
            res = f.read(step)
            if line not in res:
                errors.append("Expected: {}\nGot: {}".format(expected, res))

    @staticmethod
    def mock_complex_consumer(f, expected, n_steps, errors):
        for n in range(n_steps):
            f.read()
        res = f.readline() + f.read(10)
        if res != expected:
            errors.append("Expected: {}\nGot: {}".format(expected, res))

    def phout_multi_read(self):
        with open(self.filename) as f:
            exp = f.read()
        errors = []
        stop = Event()
        mr = FileMultiReader(self.filename, stop)
        threads = [Thread(target=self.mock_consumer,
                          args=(mr.get_file(i), exp, i, errors),
                          name='Thread-%d' % i) for i in [1000, 4000, 8000]]
        [th.start() for th in threads]
        stop.set()
        [th.join() for th in threads]
        mr.close()
        return errors

    def phout_multi_readline(self):
        errors = []
        stop = Event()
        mr = FileMultiReader(self.filename, stop)
        threads = [Thread(target=self.mock_complex_consumer,
                          args=(mr.get_file(i), exp, 10, errors),
                          name='Thread-%d' % i) for i, exp in
                   [(1000, '\n1543699431'),
                    (4000, '815\t0\t200\n1543699487'),
                    (8000, '10968\t3633\t16\t7283\t36\t7387\t1066\t328\t0\t405\n1543699534')]]
        [th.start() for th in threads]
        stop.set()
        [th.join() for th in threads]
        mr.close()
        return errors

    @pytest.mark.skip('no module in arcadia')
    @pytest.mark.benchmark(min_rounds=10)
    def test_read(self, benchmark):
        errors = benchmark(self.phout_multi_read)
        assert len(errors) == 0

    @pytest.mark.skip('no module in arcadia')
    @pytest.mark.benchmark(min_rounds=5)
    def test_readline(self, benchmark):
        errors = benchmark(self.phout_multi_readline)
        assert len(errors) == 0


class TestSecuredShell(object):

    def test_ssh_path(self):
        s = SecuredShell(None, None, None, command_timeout=10, ssh_key_path=".")
        assert s.key_filename is not None



import logging
import re
from configparser import RawConfigParser, ParsingError
from functools import reduce

import pkg_resources
import yaml

from yandextank.common.util import recursive_dict_update
from yandextank.validator.validator import load_plugin_schema, load_yaml_schema

logger = logging.getLogger(__name__)
CORE_SCHEMA = load_yaml_schema(pkg_resources.resource_filename('yandextank.core', 'config/schema.yaml'))['core']['schema']

DEPRECATED_SECTIONS = ['lunaport', 'aggregator']


def old_plugin_mapper(package):
    MAP = {'Overload': 'DataUploader'}
    return MAP.get(package, package)


def parse_package_name(package_path):
    if package_path.startswith("Tank/Plugins/"):
        package = package_path.split('/')[-1].split('.')[0]
    else:
        package = package_path.split('.')[-1].split()[0]
    return old_plugin_mapper(package)


SECTIONS_PATTERNS = {
    'tank': 'core|tank',
    'Aggregator': 'aggregator',
    'Android': 'android',
    'Appium': 'appium',
    'Autostop': 'autostop',
    'BatteryHistorian': 'battery_historian',
    'Bfg': 'bfg|ultimate_gun|http_gun|custom_gun|scenario_gun',
    'Phantom': 'phantom(-.*)?',
    'DataUploader': 'meta|overload|uploader|datauploader',
    'Telegraf': 'telegraf|monitoring',
    'JMeter': 'jmeter',
    'ResourceCheck': 'rcheck',
    'ShellExec': 'shell_?exec',
    'ShootExec': 'shoot_?exec',
    'Console': 'console',
    'TipsAndTricks': 'tips',
    'RCAssert': 'rcassert',
    'JsonReport': 'json_report|jsonreport',
    'Pandora': 'pandora',
    'InfluxUploader': 'influx',
    'OpenTSDBUploader': 'opentsdb',

}


class ConversionError(Exception):
    MSG = 'ConversionError:\n{}\nCheck your file format'

    def __init__(self, message=''):
        self.message = self.MSG.format(message)


class OptionsConflict(ConversionError):
    pass


class UnrecognizedSection(ConversionError):
    pass


def guess_plugin(section):
    for plugin, section_name_pattern in SECTIONS_PATTERNS.items():
        if re.match(section_name_pattern, section):
            return plugin
    else:
        raise UnrecognizedSection('Section {} did not match any plugin'.format(section))


def convert_rps_schedule(key, value):
    return {'load_profile': {
        'load_type': 'rps',
        'schedule': value
    }}


def convert_instances_schedule(key, value):
    return {'load_profile': {
        'load_type': 'instances',
        'schedule': value
    }}


def convert_stpd_schedule(key, value):
    return {'load_profile': {
        'load_type': 'stpd_file',
        'schedule': value
    }}


def to_bool(value):
    try:
        return bool(int(value))
    except ValueError:
        return True if 'true' == value.lower() else False


def is_option_deprecated(plugin, option_name):
    DEPRECATED = {
        'Aggregator': [
            'time_periods',
            'precise_cumulative'
        ],
        'DataUploader': [
            'copy_config_to'
        ]
    }
    if option_name in DEPRECATED.get(plugin, []):
        logger.warning('Deprecated option %s in plugin %s, omitting', option_name, plugin)
        return True
    else:
        return False


def check_options(plugin, options):
    CONFLICT_OPTS = {
        'Phantom': [{'rps_schedule', 'instances_schedule', 'stpd_file'}]
    }
    for conflict_options in CONFLICT_OPTS.get(plugin, []):
        intersect = {option[0] for option in options} & conflict_options
        if len(intersect) > 1:
            raise OptionsConflict('Conflicting options: {}: {}'.format(plugin, list(intersect)))
    return plugin, options


def without_deprecated(plugin, options):
    """
    :type options: list of tuple
    """
    return filter(lambda option: not is_option_deprecated(plugin, option[0]), options)


def old_section_name_mapper(name):
    MAP = {
        'monitoring': 'telegraf',
    }
    return MAP.get(name, name)


def rename(name):
    MAP = {
        'meta': 'uploader'
    }
    return MAP.get(name, name)


class Package(object):
    def __init__(self, package_path):
        if package_path.startswith("Tank/Plugins/"):
            self.package = package_path.split('.')[0].replace('Tank/Plugins/', 'yandextank.plugins.')
        else:
            self.package = package_path
        self.plugin_name = old_plugin_mapper(self.package.split('.')[-1])


class UnknownOption(ConversionError):

    def __init__(self, option):
        self.message = 'Unknown option: {}'.format(option)


def empty_to_none(func):
    def new_func(k, v):
        if v in '':
            return {k: None}
        else:
            return func(k, v)
    return new_func


class Option(object):
    TYPE_CASTERS = {
        'boolean': empty_to_none(lambda k, v: {k: to_bool(v)}),
        'integer': empty_to_none(lambda k, v: {k: int(v)}),
        'list': empty_to_none(lambda k, v: {k: [_.strip() for _ in v.strip().split()]}),
        'float': empty_to_none(lambda k, v: {k: float(v)})
    }

    SPECIAL_CONVERTERS = {
        'Phantom': {
            'rps_schedule': convert_rps_schedule,
            'instances_schedule': convert_instances_schedule,
            'stpd_file': convert_stpd_schedule,
            'autocases': TYPE_CASTERS['integer'],
            'headers': lambda key, value: {key: re.compile(r'\[(.*?)\]').findall(value)}
        },
        'Bfg': {
            'rps_schedule': convert_rps_schedule,
            'instances_schedule': convert_instances_schedule,
            'headers': lambda key, value: {key: re.compile(r'\[(.*?)\]').findall(value)}
        },
        'JMeter': {
            'exclude_markers': lambda key, value: {key: value.strip().split(' ')}
        },
        'Pandora': {
            'config_content': lambda key, value: {key: yaml.load(value, Loader=yaml.FullLoader)}  # works for json as well
        },
        'Autostop': {
            'autostop': lambda k, v: {k: re.findall(r'\w+\(.+?\)', v)}
        },
        'DataUploader': {
            'lock_targets': lambda k, v: {k: v.strip().split() if v != 'auto' else v}
        },
        'core': {
            'ignore_locks': lambda k, v: {'ignore_lock': to_bool(v)}
        }
    }
    CONVERTERS_FOR_UNKNOWN = {
        'DataUploader': lambda k, v: {'meta': {k: v}},
        'JMeter': lambda k, v: {'variables': {k: v}}
    }

    def __init__(self, plugin_name, key, value, schema=None):
        self.dummy_converter = lambda k, v: {k: v}
        self.plugin = plugin_name
        self._schema = schema

        if '.' in key:
            self.name, rest = key.split('.', 1)
            self.value = Option(plugin_name, rest, value, schema=self.schema[self.name]).converted
            self._converter = self.dummy_converter
        else:
            self.name = key
            self.value = value
            self._converter = None
        self._converted = None
        self._as_tuple = None

    @property
    def schema(self):
        if self._schema is None:
            module_paths = {
                'tank': 'yandextank.core'
            }

            def default_path(plugin):
                'yandextank.plugins.{}'.format(plugin)

            self._schema = load_plugin_schema(module_paths.get(self.plugin, default_path(self.plugin)))
        return self._schema

    @property
    def converted(self):
        """
        :rtype: {str: object}
        """
        if self._converted is None:
            self._converted = self.converter(self.name, self.value)
        return self._converted

    @property
    def as_tuple(self):
        """
        :rtype: (str, object)
        """
        if self._as_tuple is None:
            self._as_tuple = next(iter(self.converted.items()))
        return self._as_tuple

    @property
    def converter(self):
        """
        :rtype: callable
        """
        if self._converter is None:
            try:
                self._converter = self.SPECIAL_CONVERTERS[self.plugin][self.name]
            except KeyError:
                try:
                    self._converter = self._get_scheme_converter()
                except UnknownOption:
                    self._converter = self.CONVERTERS_FOR_UNKNOWN.get(self.plugin, self.dummy_converter)
        return self._converter

    def _get_scheme_converter(self):
        if self.name == 'enabled':
            return self.TYPE_CASTERS['boolean']
        if self.schema.get(self.name) is None:
            logger.warning('Unknown plugin option %s:%s', self.plugin, self.name)
            raise UnknownOption('{}:{}'.format(self.plugin, self.name))

        _type = self.schema[self.name].get('type', None)
        if _type is None:
            logger.warning('Option %s:%s: no type specified in schema', self.plugin, self.name)
            return self.dummy_converter
        if not isinstance(_type, str):
            # schemas 'type' field is not limited to string values
            return self.dummy_converter
        return self.TYPE_CASTERS.get(_type, self.dummy_converter)


class Section(object):
    def __init__(self, name, plugin, options, enabled=None):
        self.name = old_section_name_mapper(name)
        self.new_name = rename(self.name)
        self.plugin = plugin
        self._schema = None
        self.options = [Option(plugin, *option, schema=self.schema) for option in without_deprecated(*check_options(plugin, options))]
        self.enabled = enabled
        self._merged_options = None

    @property
    def schema(self):
        if self._schema is None:
            self._schema = load_plugin_schema('yandextank.plugins.' + self.plugin)
        return self._schema

    def get_cfg_dict(self, with_meta=True):
        options_dict = self.merged_options
        if with_meta:
            if self.plugin:
                options_dict.update({'package': 'yandextank.plugins.{}'.format(self.plugin)})
            if self.enabled is not None:
                options_dict.update({'enabled': self.enabled})
        return options_dict

    @property
    def merged_options(self):
        if self._merged_options is None:
            self._merged_options = reduce(lambda acc, upd: recursive_dict_update(acc, upd),
                                          [opt.converted for opt in self.options],
                                          {})
        return self._merged_options

    @classmethod
    def from_multiple(cls, sections, parent_name=None, child_name=None, is_list=True):
        """
        :type parent_name: str
        :type sections: list of Section
        """
        if len(sections) == 1:
            return sections[0]
        if parent_name:
            master_section = next(filter(lambda section: section.name == parent_name, sections))
            rest = filter(lambda section: section is not master_section, sections)
        else:
            master_section = sections[0]
            parent_name = master_section.name
            rest = sections[1:]
        child = {'multi': [section.get_cfg_dict(with_meta=False) for section in rest]} if is_list \
            else {child_name: cls._select_one(master_section, rest).get_cfg_dict(with_meta=False)}
        master_section.merged_options.update(child)
        return master_section

    def __repr__(self):
        return '{}/{}'.format(self.name, self.plugin)

    @classmethod
    def _select_one(cls, master_section, rest):
        MAP = {
            'bfg': lambda section: section.name == '{}_gun'.format(master_section.get_cfg_dict()['gun_type'])
        }
        return next(filter(MAP.get(master_section.name, lambda x: True), rest))
        # return filter(lambda section: section.name == MAP.get(master_section.name, ), rest)[0]


def without_defaults(cfg_ini, section):
    """

    :rtype: (str, str)
    :type cfg_ini: ConfigParser
    """
    defaults = cfg_ini.defaults()
    options = cfg_ini.items(section) if cfg_ini.has_section(section) else []
    return [(key, value) for key, value in options if key not in defaults]


PLUGIN_PREFIX = 'plugin_'
CORE_SECTION_PATTERN = 'tank|core'
CORE_SECTION_OLD = 'tank'
CORE_SECTION_NEW = 'core'


def parse_sections(cfg_ini):
    """
    :type cfg_ini: ConfigParser
    """
    return [Section(section.lower(),
                    guess_plugin(section.lower()),
                    without_defaults(cfg_ini, section))
            for section in cfg_ini.sections()
            if not re.match(CORE_SECTION_PATTERN, section.lower()) and section.lower() not in DEPRECATED_SECTIONS]


class PluginInstance(object):
    def __init__(self, name, package_and_section):
        self.name = name
        self.enabled = len(package_and_section) > 0
        try:
            package_path, self.section_name = package_and_section.split()
            self.package = Package(package_path)
        except ValueError:
            self.package = Package(package_and_section)
            self.section_name = self._guess_section_name()
        self.plugin_name = self.package.plugin_name

    def __repr__(self):
        return self.name

    def _guess_section_name(self):
        package_map = {
            'Aggregator': 'aggregator',
            'Autostop': 'autostop',
            'BatteryHistorian': 'battery_historian',
            'Bfg': 'bfg',
            'Console': 'console',
            'DataUploader': 'meta',
            'JMeter': 'jmeter',
            'JsonReport': 'json_report',
            'Maven': 'maven',
            'Monitoring': 'monitoring',
            'Pandora': 'pandora',
            'Phantom': 'phantom',
            'RCAssert': 'rcassert',
            'ResourceCheck': 'rcheck',
            'ShellExec': 'shellexec',
            'ShootExec': 'shootexec',
            'SvgReport': 'svgreport',
            'Telegraf': 'telegraf',
            'TipsAndTricks': 'tips'
        }
        name_map = {
            'aggregate': 'aggregator',
            'overload': 'overload',
            'jsonreport': 'json_report'
        }
        return name_map.get(self.name, package_map.get(self.package.plugin_name, self.name))


def enable_sections(sections, core_opts):
    """

    :type sections: list of Section
    """
    DEPRECATED_PLUGINS = ['yandextank.plugins.Aggregator', 'Tank/Plugins/Aggregator.py']

    plugin_instances = [PluginInstance(key.split('_')[1], value) for key, value in core_opts if
                        key.startswith(PLUGIN_PREFIX) and value not in DEPRECATED_PLUGINS]
    enabled_instances = {instance.section_name: instance for instance in plugin_instances if instance.enabled}
    disabled_instances = {instance.section_name: instance for instance in plugin_instances if not instance.enabled}

    for section in sections:
        if enabled_instances.pop(section.name, None) is not None:
            section.enabled = True
        elif disabled_instances.pop(section.name, None) is not None:
            section.enabled = False
    # add leftovers
    leftovers = set(enabled_instances.keys()) | set(disabled_instances.keys())
    for plugin_instance in filter(
        lambda lo: lo.section_name in leftovers,
        plugin_instances,
    ):
        sections.append(Section(plugin_instance.section_name, plugin_instance.plugin_name, [], plugin_instance.enabled))
    return sections


def combine_sections(sections):
    """
    :type sections: list of Section
    :rtype: list of Section
    """
    PLUGINS_TO_COMBINE = {
        'Phantom': ('phantom', 'multi', True),
        'Bfg': ('bfg', 'gun_config', False)
    }
    plugins = {}
    ready_sections = []
    for section in sections:
        if section.plugin in PLUGINS_TO_COMBINE:
            try:
                plugins[section.plugin].append(section)
            except KeyError:
                plugins[section.plugin] = [section]
        else:
            ready_sections.append(section)

    for plugin_name, _sections in plugins.items():
        if isinstance(_sections, list):
            parent_name, child_name, is_list = PLUGINS_TO_COMBINE[plugin_name]
            ready_sections.append(Section.from_multiple(_sections, parent_name, child_name, is_list))
    return ready_sections


def core_options(cfg_ini):
    return cfg_ini.items(CORE_SECTION_OLD) if cfg_ini.has_section(CORE_SECTION_OLD) else []


def convert_ini(ini_file):
    cfg_ini = RawConfigParser(strict=False)
    try:
        if isinstance(ini_file, str):
            cfg_ini.read(ini_file)
        else:
            cfg_ini.read_file(ini_file)
    except ParsingError as e:
        raise ConversionError(str(e))

    ready_sections = enable_sections(combine_sections(parse_sections(cfg_ini)), core_options(cfg_ini))

    plugins_cfg_dict = {section.new_name: section.get_cfg_dict() for section in ready_sections}

    plugins_cfg_dict.update({
        'core': dict([Option('core', key, value, CORE_SCHEMA).as_tuple
                      for key, value in without_defaults(cfg_ini, CORE_SECTION_OLD)
                      if not key.startswith(PLUGIN_PREFIX)])
    })
    logger.info('Converted config:\n%s', yaml.dump(plugins_cfg_dict))
    return plugins_cfg_dict


def convert_single_option(key, value):
    """

    :type value: str
    :type key: str
    :rtype: {str: obj}
    """
    section_name, option_name = key.strip().split('.', 1)
    if not re.match(CORE_SECTION_PATTERN, section_name):
        section = Section(section_name,
                          guess_plugin(section_name),
                          [(option_name, value)])
        return {section.new_name: section.get_cfg_dict()}
    else:
        if option_name.startswith(PLUGIN_PREFIX):
            return {section.new_name: section.get_cfg_dict() for section in enable_sections([], [(option_name, value)])}
        else:
            return {'core': Option('core', option_name, value, CORE_SCHEMA).converted}
